# prevent circular imports
from __future__ import annotations
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from .NodeGUI import NodeGUI
    from .NodeItem import NodeItem
    from ..FlowView import FlowView

from qtpy.QtCore import QPointF, QRectF, Qt, QSizeF
from qtpy.QtWidgets import QGraphicsWidget, QGraphicsLinearLayout, QSizePolicy, QWidget

from typing import Optional

from .NodeItem_CollapseButton import NodeItem_CollapseButton
from ..FlowViewProxyWidget import FlowViewProxyWidget
# from .Node import Node
from .NodeItem_Icon import NodeItem_Icon
from .NodeItem_TitleLabel import TitleLabel
from .PortItem import InputPortItem, OutputPortItem

from ryvencore import Flow


class NodeItemWidget(QGraphicsWidget):
    """The QGraphicsWidget managing all GUI components of a NodeItem in widgets and layouts."""

    def __init__(self, node_gui: NodeGUI, node_item: NodeItem) -> None:
        super().__init__(parent=node_item)

        self.node_gui: NodeGUI = node_gui
        self.node_item: NodeItem = node_item
        self.flow_view: FlowView = self.node_item.flow_view
        self.flow: Flow = self.flow_view.flow

        self.body_padding = 6
        self.header_padding = (0, 0, 0, 0)  # theme dependent and hence updated in setup_layout()!

        self.icon = NodeItem_Icon(node_gui, node_item) if node_gui.icon else None
        self.title_label = TitleLabel(node_gui, node_item)
        self.main_widget_proxy: Optional[FlowViewProxyWidget] = None
        if self.node_item.main_widget:
            assert isinstance(self.node_item.main_widget, QWidget)
            self.main_widget_proxy = FlowViewProxyWidget(self.flow_view)
            self.main_widget_proxy.setWidget(self.node_item.main_widget)
        self.header_layout: Optional[QGraphicsWidget]
        self.header_widget: Optional[QGraphicsWidget]
        self.collapse_button: Optional[NodeItem_CollapseButton]
        self.body_layout: QGraphicsLinearLayout
        self.body_widget: QGraphicsWidget
        self.inputs_layout: QGraphicsLinearLayout
        self.outputs_layout: QGraphicsLinearLayout
        self.setLayout(self.setup_layout())

    def setup_layout(self) -> QGraphicsLinearLayout:

        self.header_padding = self.node_item.session_design.flow_theme.header_padding

        #   main layout
        layout = QGraphicsLinearLayout(Qt.Vertical)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setSpacing(0)

        if self.node_gui.style == 'normal':
            self.collapse_button = NodeItem_CollapseButton(self.node_gui, self.node_item)

            self.header_widget = QGraphicsWidget()
            # self.header_widget.setContentsMargins(0, 0, 0, 0)
            self.header_widget.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
            self.header_layout = QGraphicsLinearLayout(Qt.Horizontal)
            self.header_layout.setSpacing(5)
            self.header_layout.setContentsMargins(
                *self.header_padding
            )
            if self.icon:
                self.header_layout.addItem(self.icon)
                self.header_layout.setAlignment(self.icon, Qt.AlignVCenter | Qt.AlignLeft)

            self.header_layout.addItem(self.title_label)

            self.header_layout.addItem(self.collapse_button)
            self.header_layout.setAlignment(self.collapse_button, Qt.AlignVCenter | Qt.AlignRight)

            self.header_widget.setLayout(self.header_layout)
            # layout.addItem(self.header_layout)
            layout.addItem(self.header_widget)
            # layout.setAlignment(self.title_label, Qt.AlignTop)
        else:
            self.collapse_button = None
            self.header_widget = None
            self.header_layout = None
            self.setZValue(self.title_label.zValue() + 1)

        #   inputs
        self.inputs_layout = QGraphicsLinearLayout(Qt.Vertical)
        self.inputs_layout.setSpacing(2)

        #   outputs
        self.outputs_layout = QGraphicsLinearLayout(Qt.Vertical)
        self.outputs_layout.setSpacing(2)

        #   body
        self.body_widget = QGraphicsWidget()
        # self.body_widget.setContentsMargins(0, 0, 0, 0)
        self.body_layout = QGraphicsLinearLayout(Qt.Horizontal)
        self.body_layout.setContentsMargins(
            self.body_padding,
            self.body_padding,
            self.body_padding,
            self.body_padding
        )

        self.body_layout.setSpacing(4)
        self.body_layout.addItem(self.inputs_layout)
        self.body_layout.setAlignment(self.inputs_layout, Qt.AlignVCenter | Qt.AlignLeft)
        self.body_layout.addStretch()
        self.body_layout.addItem(self.outputs_layout)
        self.body_layout.setAlignment(self.outputs_layout, Qt.AlignVCenter | Qt.AlignRight)

        self.body_widget.setLayout(self.body_layout)

        # layout.addItem(self.body_layout)
        layout.addItem(self.body_widget)

        return layout

    def rebuild_ui(self):
        """Due to some really strange and annoying behaviour of these QGraphicsWidgets, they don't want to shrink
        automatically when content is removed, they just stay large, even with a Minimum SizePolicy. I didn't find a
        way around that yet, so for now I have to recreate the whole layout and make sure the widget uses the smallest
        size possible."""

        # if I don't manually remove the ports from the layouts,
        # they get deleted when setting the widget's layout to None below
        for i, inp in enumerate(self.node_item.inputs):
            self.inputs_layout.removeAt(0)
            # inp.rebuild_ui()
        for i, out in enumerate(self.node_item.outputs):
            self.outputs_layout.removeAt(0)
            # out.rebuild_ui()

        self.setLayout(None)
        self.resize(self.minimumSize())
        self.setLayout(self.setup_layout())

        if self.node_item.collapsed:
            return

        for inp_item in self.node_item.inputs:
            self.add_input_to_layout(inp_item)
        for out_item in self.node_item.outputs:
            self.add_output_to_layout(out_item)

        if self.node_item.main_widget:
            self.add_main_widget_to_layout()

    def update_shape(self):

        self.title_label.update_shape()

        # makes extended node items shrink according to resizing input widgets
        if not self.node_item.initializing:
            self.rebuild_ui()
        # strangely, this only works for small node items without this, not for normal ones

        mw = self.node_item.main_widget
        if mw is not None:  # maybe the main_widget got resized
            assert self.main_widget_proxy is not None

            # self.main_widget_proxy.setMaximumSize(mw.size())

            # self.main_widget_proxy.setMaximumSize(mw.maximumSize())

            self.main_widget_proxy.setMaximumSize(QSizeF(mw.size()))
            self.main_widget_proxy.setMinimumSize(QSizeF(mw.size()))

            self.adjustSize()
            self.adjustSize()

        self.body_layout.invalidate()
        self.layout().invalidate()
        self.layout().activate()
        # very essential; repositions everything in case content has changed (inputs/outputs/widget)

        if self.node_gui.style == 'small':

            # making it recompute its true minimumWidth here
            self.adjustSize()

            if self.layout().minimumWidth() < self.title_label.width + 15:
                self.layout().setMinimumWidth(self.title_label.width + 15)
                self.layout().activate()

        w = self.boundingRect().width()
        h = self.boundingRect().height()
        rect = QRectF(QPointF(-w / 2, -h / 2),
                      QPointF(w / 2, h / 2))
        self.setPos(rect.left(), rect.top())

        if not self.node_gui.style == 'normal':
            if self.icon:
                self.icon.setPos(
                    QPointF(-self.icon.boundingRect().width() / 2,
                            -self.icon.boundingRect().height() / 2)
                )
                self.title_label.hide()
            else:
                self.title_label.setPos(
                    QPointF(-self.title_label.boundingRect().width() / 2,
                            -self.title_label.boundingRect().height() / 2)
                )


    def add_main_widget_to_layout(self):
        assert self.main_widget_proxy is not None
        if self.node_gui.main_widget_pos == 'between ports':
            self.body_layout.insertItem(1, self.main_widget_proxy)
            self.body_layout.insertStretch(2)

        elif self.node_gui.main_widget_pos == 'below ports':
            self.layout().addItem(self.main_widget_proxy)
            self.layout().setAlignment(self.main_widget_proxy, Qt.AlignHCenter)

    def add_input_to_layout(self, inp: InputPortItem):
        if self.inputs_layout.count() > 0:
            self.inputs_layout.addStretch()
        self.inputs_layout.addItem(inp)
        self.inputs_layout.setAlignment(inp, Qt.AlignLeft)

    def insert_input_into_layout(self, index: int, inp: InputPortItem):
        self.inputs_layout.insertItem(index * 2 + 1, inp)  # *2 bcs of the stretches
        self.inputs_layout.setAlignment(inp, Qt.AlignLeft)
        if len(self.node_gui.node.inputs) > 1:
            self.inputs_layout.insertStretch(index * 2 + 1)  # *2+1 because of the stretches, too

    def remove_input_from_layout(self, inp: InputPortItem):
        self.inputs_layout.removeItem(inp)

        # just a temporary workaround for the issues discussed here:
        # https://forum.qt.io/topic/116268/qgraphicslayout-not-properly-resizing-to-change-of-content
        self.rebuild_ui()

    def add_output_to_layout(self, out: OutputPortItem):
        if self.outputs_layout.count() > 0:
            self.outputs_layout.addStretch()
        self.outputs_layout.addItem(out)
        self.outputs_layout.setAlignment(out, Qt.AlignRight)

    def insert_output_into_layout(self, index: int, out: OutputPortItem):
        self.outputs_layout.insertItem(index * 2 + 1, out)  # *2 because of the stretches
        self.outputs_layout.setAlignment(out, Qt.AlignRight)
        if len(self.node_gui.node.outputs) > 1:
            self.outputs_layout.insertStretch(index * 2 + 1)  # *2+1 because of the stretches, too

    def remove_output_from_layout(self, out: OutputPortItem):
        self.outputs_layout.removeItem(out)

        # just a temporary workaround for the issues discussed here:
        # https://forum.qt.io/topic/116268/qgraphicslayout-not-properly-resizing-to-change-of-content
        self.rebuild_ui()

    def collapse(self):
        self.body_widget.hide()
        if self.main_widget_proxy is not None:
            self.main_widget_proxy.hide()

    def expand(self):
        self.body_widget.show()
        if self.main_widget_proxy is not None:
            self.main_widget_proxy.show()

    def hide_unconnected_ports(self):
        for inp in self.node_item.node.inputs:
            if self.flow.connected_output(inp) is None:
                inp.hide()
        for out in self.node_item.node.outputs:
            if len(self.flow.connected_inputs(out)):
                out.hide()

    def show_unconnected_ports(self):
        for inp in self.node_item.inputs:
            inp.show()
        for out in self.node_item.outputs:
            out.show()



# prevent circular imports
from __future__ import annotations
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from ..FlowView import FlowView

import traceback
from typing import Optional, Tuple, List

from qtpy.QtWidgets import QGraphicsItem, QGraphicsObject, QMenu, QGraphicsDropShadowEffect
from qtpy.QtCore import Qt, QRectF, QObject, QPointF
from qtpy.QtGui import QColor

from .NodeErrorIndicator import NodeErrorIndicator
from .NodeGUI import NodeGUI
from ...GUIBase import GUIBase
from .NodeItemAction import NodeItemAction
from .NodeItemAnimator import NodeItemAnimator
from .NodeItemWidget import NodeItemWidget
from .PortItem import InputPortItem, OutputPortItem
from ...utils import serialize, deserialize, MovementEnum, generate_name

from ryvencore_qt.src.Design import Design

from ryvencore import Node
from ryvencore.NodePort import NodeInput, NodeOutput


class NodeItem(GUIBase, QGraphicsObject):  # QGraphicsItem, QObject):
    """The GUI representative for nodes. Unlike the Node class, this class is not subclassed individually and works
    the same for every node."""

    def __init__(self, node: Node, node_gui: NodeGUI, flow_view: FlowView, design: Design):
        GUIBase.__init__(self, representing_component=node)
        QGraphicsObject.__init__(self)

        self.node = node
        self.node_gui = node_gui
        self.node_gui.item = self
        self.flow_view = flow_view
        self.session_design: Design = design
        self.movement_state: Optional[MovementEnum] = None
        self.movement_pos_from: Optional[QPointF] = None
        self.painted_once: bool = False
        self.inputs: List[InputPortItem] = []
        self.outputs: List[OutputPortItem] = []
        self.color = QColor(self.node_gui.color)  # manipulated by self.animator

        self.collapsed = False
        self.hovered = False
        self.hiding_unconnected_ports = False
        self.displaying_error = False

        # 'initializing' will be set to False below. It's needed for the ports setup, to prevent shape updating stuff
        self.initializing = True

        # self.temp_state_data = None
        self.init_data = self.node.load_data

        # CONNECT TO NODE
        self.node_gui.updating.connect(self.node_updating)
        self.node_gui.update_shape_triggered.connect(self.update_shape)
        self.node_gui.hide_unconnected_ports_triggered.connect(self.hide_unconnected_ports_triggered)
        self.node_gui.show_unconnected_ports_triggered.connect(self.show_unconnected_ports_triggered)
        self.node_gui.input_added.connect(self.on_node_input_added)
        self.node_gui.output_added.connect(self.on_node_output_added)
        self.node_gui.input_removed.connect(self.on_node_input_removed)
        self.node_gui.output_removed.connect(self.on_node_output_removed)

        # FLAGS
        self.setFlags(
            QGraphicsItem.ItemIsSelectable |
            QGraphicsItem.ItemIsMovable |
            QGraphicsItem.ItemSendsScenePositionChanges
        )

        self.setAcceptHoverEvents(True)
        self.setCacheMode(QGraphicsItem.DeviceCoordinateCache)

        # UI
        self.shadow_effect = None
        self.main_widget = None
        if self.node_gui.main_widget_class is not None:
            self.main_widget = self.node_gui.main_widget_class((self.node, self, self.node_gui))
        self.widget = NodeItemWidget(self.node_gui, self)  # QGraphicsWidget(self)
        self.animator = NodeItemAnimator(self)  # needs self.title_label
        self.error_indicator = NodeErrorIndicator(self)
        self.error_indicator.hide()

        # TOOLTIP
        self.tooltip_descr_html_content = \
            self.node_gui.description_html \
            if self.node_gui.description_html is not None \
            else \
            f'<p>{self.node.__doc__}</p>'

        self.setToolTip(f'<html><head/><body>{self.tooltip_descr_html_content}</body></html>')

        # DESIGN THEME
        self.session_design.flow_theme_changed.connect(self.update_design)
        self.session_design.performance_mode_changed.connect(self.update_design)

    def initialize(self):
        """All ports and the main widget get finally created here."""

        # LOADING DATA
        if self.init_data is not None:
            if self.main_widget:
                try:
                    self.main_widget.set_state(deserialize(self.init_data['main widget data']))
                except Exception as e:
                    print('Exception while setting data in', self.node.title, 'Node\'s main widget:', e,
                          ' (was this intended?)')

        # catch up on init ports
        for inp in self.node.inputs:
            self.add_new_input(inp)

        for out in self.node.outputs:
            self.add_new_output(out)

        if self.init_data is not None:
            if self.init_data.get('unconnected ports hidden'):
                self.hide_unconnected_ports_triggered()
            if self.init_data.get('collapsed'):
                self.collapse()

        if self.init_data is not None:
            self.node_gui.load(self.init_data)

        self.node_gui.initialized()

        self.initializing = False

        # No self.update_shape() here because for some reason, the bounding rect hasn't been initialized yet, so
        # self.update_shape() gets called when the item is being drawn the first time (see paint event in NI painter)
        # https://forum.qt.io/topic/117179/force-qgraphicsitem-to-update-immediately-wait-for-update-event

        self.update_design()  # load current design, update QGraphicsItem

        self.update()  # ... not sure if I need that

    def __str__(self):
        name = self.node.__class__.title if self.node else f'{NodeItem.__name__}'
        obj = self.node if self.node else self
        return generate_name(obj, name)
    
    # UI STUFF

    def node_updating(self):
        if self.session_design.animations_enabled:
            if not self.animator.running():
                self.animator.start()
            elif self.animator.fading_out():
                self.animator.set_animation_max()

        self.update()

    def display_error(self, e):
        self.error_indicator.set_error(e)
        self.error_indicator.show()
        self.displaying_error = True

    def remove_error_message(self):
        self.error_indicator.hide()
        self.setToolTip(f'<html><head/><body>{self.tooltip_descr_html_content}</body></html>')

    def set_tooltip(self, error_msg=None):

        if error_msg is not None:
            err = f'<p style="background: red; color: white">{error_msg}</p>'
        else:
            err = ''

        if self.node.description_html:
            html = self.node.description_html + f'<html><head/><body>{err}</body></html>'
        elif self.node.__doc__:
            html = f'<html><head/><body><p>{self.node.__doc__}</p>{err}</body></html>'

        self.setToolTip(html)
        self.setCursor(Qt.SizeAllCursor)

    def on_node_input_added(self, index, inp: NodeInput):
        insert = index if index == len(self.node.inputs) - 1 else None
        self.add_new_input(inp, insert)

    def add_new_input(self, inp: NodeInput, insert: Optional[int] = None):

        if inp in self.node_gui.input_widgets:
            widget_name = self.node_gui.input_widgets[inp]['name']
            widget_class = self.node_gui.input_widget_classes[widget_name]
            widget_pos = self.node_gui.input_widgets[inp]['pos']
            widget = (widget_class, widget_pos)
        else:
            widget = None

        # create item
        item = InputPortItem(self.node_gui, self, inp, input_widget=widget)

        if insert is not None:
            self.inputs.insert(insert, item)
            self.widget.insert_input_into_layout(insert, item)
        else:
            self.inputs.append(item)
            self.widget.add_input_to_layout(item)

        if not self.initializing:
            self.update_shape()
            self.update()

    def on_node_input_removed(self, index, inp: NodeInput):
        self.remove_input(inp)

    def remove_input(self, inp: NodeInput):
        item: InputPortItem
        for inp_item in self.inputs:
            if inp_item.port == inp:
                item = inp_item
                break
        else:
            return

        # index = self.node.inputs.index(inp)
        # item = self.inputs[index]

        # for some reason, I have to remove all widget items manually from the scene too. setting the items to
        # ownedByLayout(True) does not work, I don't know why.
        self.scene().removeItem(item.pin)
        self.scene().removeItem(item.label)
        if item.proxy is not None:
            self.scene().removeItem(item.proxy)

        self.inputs.remove(item)
        self.widget.remove_input_from_layout(item)

        if not self.initializing:
            self.update_shape()
            self.update()

    def on_node_output_added(self, index, out: NodeOutput):
        insert = index if index == len(self.node.outputs) - 1 else None
        self.add_new_output(out, insert)

    def add_new_output(self, out: NodeOutput, insert: Optional[int] = None):

        # create item
        # out.item = OutputPortItem(out.node, self, out)
        item = OutputPortItem(self.node_gui, self, out)

        if insert is not None:
            self.outputs.insert(insert, item)
            self.widget.insert_output_into_layout(insert, item)
        else:
            self.outputs.append(item)
            self.widget.add_output_to_layout(item)

        if not self.initializing:
            self.update_shape()
            self.update()

    def on_node_output_removed(self, index, out: NodeOutput):
        self.remove_output(out)

    def remove_output(self, out: NodeOutput):
        item: OutputPortItem
        for out_item in self.outputs:
            if out_item.port == out:
                item = out_item
                break
        else:
            return

        # index = self.node.outputs.index(out)
        # item = self.outputs[index]

        # see remove_input() for info!
        self.scene().removeItem(item.pin)
        self.scene().removeItem(item.label)

        self.outputs.remove(item)
        self.widget.remove_output_from_layout(item)

        if not self.initializing:
            self.update_shape()
            self.update()

    def update_shape(self):
        self.widget.update_shape()
        self.update_conn_pos()
        self.flow_view.viewport().update()

    def update_design(self):
        """Loads the shadow effect option and causes redraw with active theme."""

        if self.session_design.node_item_shadows_enabled:
            self.shadow_effect = QGraphicsDropShadowEffect()
            self.shadow_effect.setXOffset(12)
            self.shadow_effect.setYOffset(12)
            self.shadow_effect.setBlurRadius(20)
            self.shadow_effect.setColor(self.session_design.flow_theme.node_item_shadow_color)
            self.setGraphicsEffect(self.shadow_effect)
        else:
            self.setGraphicsEffect(None)

        self.widget.update_shape()
        self.animator.reload_values()

        QGraphicsItem.update(self)

    def boundingRect(self):
        # remember: (0, 0) shall be the NI's center!
        rect = QRectF()
        w = self.widget.layout().geometry().width()
        h = self.widget.layout().geometry().height()
        rect.setLeft(-w / 2)
        rect.setTop(-h / 2)
        rect.setWidth(w)
        rect.setHeight(h)
        return rect

    def get_left_body_header_vertex_scene_pos(self) -> QPointF:
        assert self.widget.header_widget is not None
        return self.mapToScene(
            QPointF(
                -self.boundingRect().width() / 2,
                -self.boundingRect().height() / 2 + self.widget.header_widget.rect().height()
            )
        )

    def get_right_body_header_vertex_scene_pos(self) -> QPointF:
        assert self.widget.header_widget is not None
        return self.mapToScene(
            QPointF(
                +self.boundingRect().width() / 2,
                -self.boundingRect().height() / 2 + self.widget.header_widget.rect().height()
            )
        )

    def hide_unconnected_ports_triggered(self):
        self.widget.hide_unconnected_ports()
        self.hiding_unconnected_ports = True
        self.update_shape()

    def show_unconnected_ports_triggered(self):
        self.widget.show_unconnected_ports()
        self.hiding_unconnected_ports = False
        self.update_shape()

    def expand(self):
        self.collapsed = False
        self.widget.expand()
        self.update_shape()

    def collapse(self):
        self.collapsed = True
        self.widget.collapse()
        self.update_shape()

    #   PAINTING

    def paint(self, painter, option, widget=None):
        """All painting is done by NodeItemPainter"""

        # in order to access a meaningful geometry of GraphicsWidget contents in update_shape(), the paint event
        # has to be called once.
        # https://forum.qt.io/topic/117179/force-qgraphicsitem-to-update-immediately-wait-for-update-event/4
        if not self.painted_once:
            # Since I am using a NodeItemWidget, calling self.update_design() here (again)
            # leads to a QT crash without error, which is strange. In pronciple, calling update_design multiple times
            # shouldn't be a problem. It's not necessary anymore, so I removed it.
            # self.update_design()

            self.update_shape()
            self.update_conn_pos()
            self.error_indicator.setPos(self.boundingRect().bottomRight())

        self.session_design.flow_theme.paint_NI(
            node_gui=self.node_gui,
            selected=self.isSelected(),
            hovered=self.hovered,
            node_style=self.node_gui.style,
            painter=painter,
            option=option,
            color=self.color,
            w=self.boundingRect().width(),
            h=self.boundingRect().height(),
            bounding_rect=self.boundingRect(),
            title_rect=self.widget.header_widget.boundingRect()
            if self.widget.header_widget
            else self.widget.title_label.boundingRect()
        )

        # useful for widget development:

        # painter.setBrush(Qt.NoBrush)
        # painter.setPen(QPen(QColor('black')))
        #
        # painter.drawRect(
        #         self.boundingRect()
        # )
        #
        # header_rect = QRectF(
        #         -self.boundingRect().width()/2, -self.boundingRect().height()/2,
        #         self.widget.header_widget.geometry().width(), self.widget.header_widget.geometry().height()
        #     )
        # painter.drawRect(header_rect)
        #
        # # note that the widget's layout has spacing 0
        # body_rect = QRectF(
        #     -self.boundingRect().width()/2, -self.boundingRect().height()/2 + header_rect.height(),
        #     self.widget.body_widget.geometry().width(), self.widget.body_widget.geometry().height()
        # )
        # painter.drawRect(body_rect)

        self.painted_once = True

    # MOUSE INTERACTION

    def get_context_menu(self):
        menu = QMenu(self.flow_view)

        actions = self.get_actions(self.node_gui.actions, menu)
        for a in actions:  # menu needed for 'parent'
            if type(a) == NodeItemAction:
                menu.addAction(a)
            elif type(a) == QMenu:
                menu.addMenu(a)

        return menu

    def itemChange(self, change, value):
        """Ensures that all connections, selection borders etc. that get drawn in the FlowView get
        constantly redrawn during a drag of the item"""

        if change == QGraphicsItem.ItemPositionChange:
            if self.session_design.performance_mode == 'pretty':
                self.flow_view.viewport().update()
            if self.movement_state == MovementEnum.mouse_clicked:
                self.movement_state = MovementEnum.position_changed

            self.update_conn_pos()

        return QGraphicsItem.itemChange(self, change, value)

    def on_move(self):
        self.update_conn_pos()
    
    def update_conn_pos(self):
        """Updates the scene positions of connections"""

        for o in self.node.outputs:
            for i in self.node.flow.connected_inputs(o):
                # c.item.recompute()

                if (o, i) not in self.flow_view.connection_items:
                    # it can happen that the connection item hasn't been
                    # created yet
                    continue

                item = self.flow_view.connection_items[(o,i)]
                item.recompute()
        for i in self.node.inputs:
            o = self.node.flow.connected_output(i)
            # c.item.recompute()

            if (o, i) not in self.flow_view.connection_items:
                # it can happen that the connection item hasn't been
                # created yet
                continue

            item = self.flow_view.connection_items[(o,i)]
            item.recompute()

    def hoverEnterEvent(self, event):
        self.hovered = True
        self.widget.title_label.set_NI_hover_state(hovering=True)
        QGraphicsItem.hoverEnterEvent(self, event)

    def hoverLeaveEvent(self, event):
        self.hovered = False
        self.widget.title_label.set_NI_hover_state(hovering=False)
        QGraphicsItem.hoverLeaveEvent(self, event)

    def mousePressEvent(self, event):
        """Used for Moving-Commands in FlowView - may be replaced later by a nicer determination of a moving action."""

        self.flow_view.mouse_event_taken = True

        if event.button() == Qt.LeftButton:
            self.movement_state = MovementEnum.mouse_clicked
            self.movement_pos_from = self.pos()
        return QGraphicsItem.mousePressEvent(self, event)

    def mouseReleaseEvent(self, event):
        """Used for Moving-Commands in FlowView - may be replaced later by a nicer determination of a moving action."""

        self.flow_view.mouse_event_taken = True

        if self.movement_state == MovementEnum.position_changed:
            self.flow_view._move_selected_copmonents__cmd(
                pos_diff=self.pos() - self.movement_pos_from,
                already_moved=True,
            )
        self.movement_state = None
        return QGraphicsItem.mouseReleaseEvent(self, event)

    # ACTIONS

    def get_actions(self, actions_dict, menu):
        actions = []

        for k in actions_dict:
            v_dict = actions_dict[k]
            try:
                method = v_dict['method']
                data = None
                try:
                    data = v_dict['data']
                except KeyError:
                    pass
                action = NodeItemAction(node_gui=self.node_gui, text=k, method=method, menu=menu, data=data)
                actions.append(action)
            except KeyError:
                action_menu = QMenu(k, menu)
                sub_actions = self.get_actions(v_dict, action_menu)
                for a in sub_actions:
                    action_menu.addAction(a)
                actions.append(action_menu)

        return actions

    # DATA

    def complete_data(self, data: dict) -> dict:
        """completes the node's data by adding all frontend info"""

        data['pos x'] = self.pos().x()
        data['pos y'] = self.pos().y()
        if self.main_widget:
            data['main widget data'] = serialize(self.main_widget.get_state())

        data['unconnected ports hidden'] = self.hiding_unconnected_ports
        data['collapsed'] = self.collapsed

        data = {**data, **self.node_gui.data()}

        return data



# prevent circular imports
from __future__ import annotations
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from ..FlowView import FlowView

from typing import Dict, Optional

from qtpy.QtWidgets import QGraphicsItem
from qtpy.QtGui import QPen, QPainter, QColor, QPainterPath
from qtpy.QtCore import Qt, QRectF, QPointF, QLineF
from ...utils import MovementEnum, generate_name

class DrawingObject(QGraphicsItem):
    """GUI implementation for 'drawing objects' in the scene, written by hand using a stylus pen"""

    def __init__(self, flow_view: FlowView, load_data: Optional[Dict] = None):
        super(DrawingObject, self).__init__()

        self.setFlags(QGraphicsItem.ItemIsSelectable | QGraphicsItem.ItemIsMovable |
                      QGraphicsItem.ItemSendsScenePositionChanges)
        self.setAcceptHoverEvents(True)
        self.setCacheMode(QGraphicsItem.DeviceCoordinateCache)  # for rendering flow pictures

        self.flow_view = flow_view
        self.color = None
        self.base_stroke_weight = None
        self.type_ = 'pen'  # so far the only available, but I already save it so I could add more types in the future
        self.points = []
        self.stroke_weights = []
        self.pen_stroke_weight = 0  # approx. avg of self.stroke_weights
        self.rect = None
        self.path: Optional[QPainterPath] = None
        self.width = -1
        self.height = -1
        self.finished = False

        # viewport_pos enables global floating points for precise pen positions
        self.viewport_pos: Optional[QPointF] = None
        # if the drawing gets loaded, its correct global floating pos is already correct (gets set by flow then)

        self.movement_state = None  # ugly - should get replaced later, see NodeItem, same issue
        self.movement_pos_from = None

        if load_data is not None: 
            if 'viewport pos' in load_data:
                self.viewport_pos = load_data['viewport pos']

            if 'points' in load_data:
                p_c = load_data['points']
                for p in p_c:
                    if type(p) == list:
                        x = p[0]
                        y = p[1]
                        w = p[2]
                        self.points.append(QPointF(x, y))
                        self.stroke_weights.append(w)
                    elif type(p) == dict:  # backwards compatibility
                        x = p['x']
                        y = p['y']
                        w = p['w']
                        self.points.append(QPointF(x, y))
                        self.stroke_weights.append(w)
                self.finished = True

            self.color = QColor(load_data['color'])
            self.base_stroke_weight = load_data['base stroke weight']

    def __str__(self):
        return generate_name(self, 'Drawing')
    
    def paint(self, painter, option, widget=None):

        if not self.finished:
            for i in range(1, len(self.points)):
                pen = QPen()
                pen.setColor(self.color)
                pen_width = (self.stroke_weights[i] + 0.2) * self.base_stroke_weight
                pen.setWidthF(pen_width)
                if i == 1 or i == len(self.points) - 1:
                    pen.setCapStyle(Qt.RoundCap)
                painter.setPen(pen)
                painter.setRenderHint(QPainter.Antialiasing)
                painter.setRenderHint(QPainter.HighQualityAntialiasing)
                painter.drawLine(self.points[i - 1], self.points[i])
            return

        if not self.path and self.finished:
            if len(self.points) == 0:
                return

            self.path = QPainterPath()
            self.path.moveTo(self.points[0])
            avg_weight = self.stroke_weights[0]
            for i in range(1, len(self.points)):
                self.path.lineTo(self.points[i])
                avg_weight += self.stroke_weights[i]
            self.pen_stroke_weight = (avg_weight/len(self.points) + 0.2)*self.base_stroke_weight

        pen = QPen()
        pen.setColor(self.color)
        pen.setWidthF(self.pen_stroke_weight)
        painter.setPen(pen)
        painter.setRenderHint(QPainter.Antialiasing)
        # painter.setRenderHint(QPainter.HighQualityAntialiasing)
        painter.drawPath(self.path)

    def append_point(self, posF_in_view: QPointF) -> bool:
        """
        Only used for active drawing.
        Appends a point (floating, in viewport coordinates) only if the distance to the last one isn't too small
        """

        p: QPointF = (self.viewport_pos + posF_in_view) - self.pos()
        p.setX(round(p.x(), 2))
        p.setY(round(p.y(), 2))

        if len(self.points) > 0:
            line = QLineF(self.points[-1], p)
            if line.length() < 0.5:
                return False

        self.points.append(p)
        return True

    def finish(self):
        """
        Computes the correct center position and updates the relative position for all points.
        """

        # Correct bounding rect (so far (0,0) is at the start of the line, but it should be in the middle)

        rect_center = self.get_points_rect_center()
        for p in self.points:
            p.setX(p.x()-rect_center.x())
            p.setY(p.y()-rect_center.y())
        self.setPos(self.pos()+rect_center)

        self.rect = self.get_points_rect()

        self.finished = True

    def get_points_rect(self):
        """Computes the 'bounding rect' for all points"""

        if len(self.points) == 0:
            return QRectF(0, 0, 0, 0)
        x_coords = [p.x() for p in self.points]
        y_coords = [p.y() for p in self.points]
        left = min(x_coords)
        right = max(x_coords)
        up = min(y_coords)
        down = max(y_coords)

        rect = QRectF(left, up, right - left, down - up)

        self.width = rect.width()
        self.height = rect.height()
        return rect

    def get_points_rect_center(self):
        """Returns the center point for the 'bounding rect' for all points"""

        return self.get_points_rect().center()

    def boundingRect(self):
        if self.rect:
            return self.rect
        else:
            return self.get_points_rect()

    def itemChange(self, change, value):
        if change == QGraphicsItem.ItemPositionChange:
            self.flow_view.viewport().update()
            if self.movement_state == MovementEnum.mouse_clicked:
                self.movement_state = MovementEnum.position_changed

        return QGraphicsItem.itemChange(self, change, value)

    def mousePressEvent(self, event):
        """Used for Moving-Commands in Flow - may be replaced later with a nicer determination of a move action."""

        self.movement_state = MovementEnum.mouse_clicked
        self.movement_pos_from = self.pos()
        return QGraphicsItem.mousePressEvent(self, event)

    def mouseReleaseEvent(self, event):
        """Used for Moving-Commands in Flow - may be replaced later with a nicer determination of a move action."""

        if self.movement_state == MovementEnum.position_changed:
            self.flow_view._move_selected_copmonents__cmd(
                pos_diff=self.pos() - self.movement_pos_from,
                already_moved=True,
            )
        self.movement_state = None
        return QGraphicsItem.mouseReleaseEvent(self, event)

    def data_(self):
        drawing_dict = {
            'pos x': self.pos().x(),
            'pos y': self.pos().y(),
            'color': self.color.name(),
            'type': self.type_,
            'base stroke weight': self.base_stroke_weight
        }
        points_list = []
        for i in range(len(self.points)):
            p = self.points[i]
            points_list.append([p.x(), p.y(), self.stroke_weights[i]])
        drawing_dict['points'] = points_list
        return drawing_dict


# alternative QGraphcsItem-based implementation:
# 
#   ...
#     def paint(self, painter, option, widget=None):
#
#         self.setBrush(Qt.NoBrush)
#
#         if self.finished:
#             super().paint(painter, option,widget)
#         else:
#             if len(self.stroke_weights) == 0:
#                 return
#
#             # pen
#             pen = QPen(self.color)
#             pen.setWidthF(self.stroke_weights[-1])
#             pen.setCapStyle(Qt.RoundCap)
#             self.setPen(pen)
#
#             # path
#             path = QPainterPath(self.points[0])
#             for i, p in enumerate(self.points, start=1):
#                 path.lineTo(p)
#             self.setPath(path)
#             self.path_generated = True
#
#             super().paint(painter, option, widget)
#
#   ...
#
#     def finish(self):
#
#         self.finished = True
#         self.update()
# 
# it is a bit worse in appearance but might be a lot faster



from typing import Any, Dict, Optional

from ryvencore.Base import Base
from qtpy.QtWidgets import QGraphicsObject, QGraphicsItem
from qtpy.QtCore import QObject

  
class GUIBase:
    """Base class for GUI items that represent specific core components"""

    # every frontend GUI object that represents some specific component from the core
    # is stored there under the the global id of the represented component.
    # used for completing data (serialization)
    FRONTEND_COMPONENT_ASSIGNMENTS: Dict[int, Any] = {}  # component global id : GUI object

    @staticmethod
    def get_complete_data_function(session):
        """
        generates a function that searches through generated data by the core and calls
        complete_data() on frontend components that represent them to add frontend data
        """

        def analyze(obj):
            """Searches recursively through obj and calls complete_data(obj) on associated
            frontend components (instances of GUIBase)"""

            if isinstance(obj, dict):
                GID = obj.get('GID')
                if GID is not None:
                    # find representative
                    comp = GUIBase.FRONTEND_COMPONENT_ASSIGNMENTS.get(GID)
                    if comp:
                        obj = comp.complete_data(obj)

                # look for child objects
                for key, value in obj.items():
                    obj[key] = analyze(value)

            elif isinstance(obj, list):
                for i in range(len(obj)):
                    item = obj[i]
                    item = analyze(item)
                    obj[i] = item

            return obj

        return analyze

    def __init__(self, representing_component: Optional[Base] = None):
        """parameter `representing` indicates representation of a specific core component"""
        if representing_component is not None:
            GUIBase.FRONTEND_COMPONENT_ASSIGNMENTS[representing_component.global_id] = self

    # OVERRIDE
    def complete_data(self, data: dict) -> dict:
        """completes the data dict of the represented core component by adding all frontend data"""
        return data
    
    def on_move(self):
        """virtual function for when a GUI is moved in the view"""
        pass

# if the __doc__ is incorrect, this class should be removed
class QGraphicsItemAnimated(QGraphicsObject):
    """
    Serves as a proxy for animating any kind fo QGraphicsItem.
    This was created because there is no apparent way to animate
    a QGraphicsItem that isn't a QObject.
    """
    
    def __init__(self, item: QGraphicsItem, parent = None) -> None:
        super().__init__(parent)
        self.item = item
        
        # for delete purposes
        # perhaps this could be implemented in an item change where the scene
        # is none and calling a delete later 
        self.item.setParentItem(self)
        self.item.setVisible(False)
    
    def boundingRect(self):
        return self.item.boundingRect()
    
    def paint(self, painter, option, widget):
        return self.item.paint(painter, option, widget)


from qtpy.QtWidgets import QVBoxLayout, QWidget, QLineEdit, QScrollArea
from qtpy.QtCore import Qt

from .VarsList_VarWidget import VarsList_VarWidget


class VariablesListWidget(QWidget):
    """Convenience class for a QWidget to easily manage script variables of a script."""

    def __init__(self, vars_addon, flow):
        super(VariablesListWidget, self).__init__()

        self.vars_addon = vars_addon
        self.flow = flow
        self.vars_addon.var_created.sub(self.on_var_created)
        self.vars_addon.var_deleted.sub(self.on_var_deleted)
        self.widgets = []
        self.currently_edited_var = ''
        self.ignore_name_line_edit_signal = False  # because disabling causes firing twice otherwise
        # self.data_type_line_edits = []  # same here

        self.setup_UI()


    def setup_UI(self):
        main_layout = QVBoxLayout()

        self.list_layout = QVBoxLayout()
        self.list_layout.setAlignment(Qt.AlignTop)

        # list scroll area

        self.list_scroll_area = QScrollArea()
        self.list_scroll_area.setWidgetResizable(True)
        self.list_scroll_area.setContentsMargins(0, 0, 0, 0)

        w = QWidget()
        w.setContentsMargins(0, 0, 0, 0)
        w.setLayout(self.list_layout)

        self.list_scroll_area.setWidget(w)

        main_layout.addWidget(self.list_scroll_area)

        # controls

        self.new_var_name_lineedit = QLineEdit()
        self.new_var_name_lineedit.setPlaceholderText('new var\'s title')
        self.new_var_name_lineedit.returnPressed.connect(self.new_var_LE_return_pressed)

        main_layout.addWidget(self.new_var_name_lineedit)

        self.setContentsMargins(0, 0, 0, 0)
        self.setLayout(main_layout)

        self.recreate_list()


    def on_var_created(self, flow, name, var):
        if flow == self.flow:
            self.widgets.append(
                VarsList_VarWidget(self, self.vars_addon, self.flow, var)
            )
            self.rebuild_list()


    def on_var_deleted(self, flow, name):
        # apparently, I cannot remove widgets the same way I add them
        self.recreate_list()


    def recreate_list(self):
        for w in self.widgets:
            w.hide()
            del w

        self.widgets.clear()
        # self.data_type_line_edits.clear()

        for var_name, var_info in self.vars_addon.flow_variables[self.flow].items():
            new_widget = VarsList_VarWidget(self, self.vars_addon, self.flow, var_info['var'])
            # new_widget.name_LE_editing_finished.connect(self.name_line_edit_editing_finished)
            self.widgets.append(new_widget)

        self.rebuild_list()


    def rebuild_list(self):
        for i in range(self.list_layout.count()):
            self.list_layout.removeItem(self.list_layout.itemAt(0))

        for w in self.widgets:
            self.list_layout.addWidget(w)


    def new_var_LE_return_pressed(self):
        name = self.new_var_name_lineedit.text()
        if not self.vars_addon.var_name_valid(self.flow, name=name):
            return
        v = self.vars_addon.create_var(self.flow, name=name)


    # def name_line_edit_editing_finished(self):
    #     var_widget: VarsList_VarWidget = self.sender()
    #     var_widget.name_line_edit.setEnabled(False)
    #
    #     # search for name issues
    #     new_var_name = var_widget.name_line_edit.text()
    #     for v in self.vars_manager.variables:
    #         if v.name == new_var_name:
    #             var_widget.name_line_edit.setText(self.currently_edited_var.name)
    #             return
    #
    #     var_widget.var.name = new_var_name


    def del_var(self, var):
        self.vars_addon.delete_var(self.flow, var.name)



from typing import List, Tuple
from enum import Enum

from qtpy.QtCore import (
    QTimeLine, 
    QPropertyAnimation, 
    QParallelAnimationGroup,
    QEasingCurve, 
)
from qtpy.QtGui import (
    QPen,
    QBrush, 
)
from qtpy.QtWidgets import (
    QGraphicsItem,
    QGraphicsObject,
)

from ...GUIBase import QGraphicsItemAnimated


class ConnPathItemsAnimation(QGraphicsObject):
    """
    Animates items over the path of a connection by updating their positions.
    When the path of the connection changes, recompute() needs to be called.
    The animation can be toggled on and off.
    """

    def __init__(
        self,
        items: List[QGraphicsItem],
        connection,
        frames=100,
        between=125,
        speed=0.15,
    ):
        super().__init__()
        
        self.items: List[QGraphicsItemAnimated] = [
            QGraphicsItemAnimated(item, self) 
            for item in items
        ]
        self.connection = connection
        self.between = between
        self.speed = speed
        self.visible_items: List[Tuple[QGraphicsItemAnimated, float]] = []
        self.__visible_flag = True
        
        self.setParentItem(self.connection)
        
        self.timeline = QTimeLine()
        self.timeline.setFrameRange(0, frames)
        self.timeline.setEasingCurve(QEasingCurve(QEasingCurve.Type.Linear))
        self.timeline.valueChanged.connect(self._update_items)
        self.timeline.setLoopCount(0)

    def boundingRect(self):
        return self.connection.boundingRect()
    
    def paint(self, painter, option, widget):
        pass
    
    def toggle(self, recompute: bool = False):
        """Toggles the path animation. Returns True if toggled on, otherwise False"""

        state = self.timeline.state()
        running = False
        if state == QTimeLine.NotRunning:
            self.timeline.start()
            running = True
        elif state == QTimeLine.Paused:
            self.timeline.resume()
            running = True
        else:
            running = False
            self.timeline.setPaused(True)
        
        if recompute:
            self.recompute()
        return running
    
    def start(self, recompute: bool = True):
        self.timeline.start()
        if recompute:
            self.recompute()

    def stop(self, recompute: bool = True):
        self.timeline.stop()
        if recompute:
            self.recompute()

    def recompute(self):
        
        if self.__visible_flag:
            for item in self.items:
                item.setVisible(False)
            self.__visible_flag = False

        if self.timeline.state() == QTimeLine.State.NotRunning:
            return
        
        self.__visible_flag = True

        path_len = self.connection.path().length()
        num_points = max(3, min(self.connection.num_dots, int(path_len / self.between)))

        self.timeline.setDuration(path_len / self.speed)
        self.visible_items = []

        for i in range(1, num_points + 1):
            percent = i / num_points
            item = self.items[i - 1]
            self.visible_items.append((item, percent))
            item.setVisible(True)
            item.item.setPen(QPen(self.connection.get_style_color(), self.connection.pen_width()))
            item.item.setBrush(QBrush(self.connection.get_style_color()))

    def _update_items(self, percent):
        for item, item_percent in self.visible_items:
            p = percent + item_percent
            if p > 1:
                p = p - 1
            item.setPos(self.connection.path().pointAtPercent(p))

    def _pause(self, paused: bool, recompute: bool = True):
        self.timeline.setPaused(paused)
        if recompute:
            self.recompute()

    def _state(self):
        return self.timeline.state()
    

class ConnPathItemsAnimationScaled:
    """
    Wraps a ConnPathItemsAnimation and adds a scale animation to it.
    """

    class State(Enum):
        NOT_RUNNING = 0
        RUNNING = 1
        TO_SCALE = 2
        TO_ZERO = 3
    
    def __init__(
        self,
        items_animation: ConnPathItemsAnimation,
        duration: int = 750, 
        scale: int = 1, 
    ) -> None:
        
        self.con_items_anim = items_animation
        self.duration = duration
        self.scalar = scale
        self.to_scalar_group = QParallelAnimationGroup()
        self.to_zero_group = QParallelAnimationGroup()
        self.state = ConnPathItemsAnimationScaled.State.NOT_RUNNING
        
        for item in self.con_items_anim.items:
            item.setScale(0)
            # to scaler
            to_scalar_anim = QPropertyAnimation(item, b'scale')  # type: ignore
            to_scalar_anim.setDuration(self.duration)
            self.to_scalar_group.addAnimation(to_scalar_anim)
            # to zero
            to_zero_anim = QPropertyAnimation(item, b'scale')  # type: ignore
            to_zero_anim.setDuration(self.duration)
            self.to_zero_group.addAnimation(to_zero_anim)

        self.to_scalar_group.finished.connect(self._on_scalar_ended)
        self.to_zero_group.finished.connect(self._on_zero_ended)
    
    def start(self):
        if (self.state == ConnPathItemsAnimationScaled.State.NOT_RUNNING or 
            self.state == ConnPathItemsAnimationScaled.State.TO_ZERO):
            self.toggle()
    
    def stop(self):
        if (self.state == ConnPathItemsAnimationScaled.State.RUNNING or
            self.state == ConnPathItemsAnimationScaled.State.TO_SCALE):
            self.toggle()
            
    def toggle(self):
        if self.state == ConnPathItemsAnimationScaled.State.NOT_RUNNING:
            self._run_scalar()
            self.con_items_anim.start()
            self.state = ConnPathItemsAnimationScaled.State.TO_SCALE
        elif self.state == ConnPathItemsAnimationScaled.State.RUNNING:
            self._run_zero()
            self.state = ConnPathItemsAnimationScaled.State.TO_ZERO
        elif self.state == ConnPathItemsAnimationScaled.State.TO_SCALE:
            self.to_scalar_group.stop()
            self._run_zero()
            self.state = ConnPathItemsAnimationScaled.State.TO_ZERO
        else:
            self.to_zero_group.stop()
            self._run_scalar()
            self.state = ConnPathItemsAnimationScaled.State.TO_SCALE
    
    def force_stop(self):
        self.to_zero_group.stop()
        self.to_scalar_group.stop()
    
    def _run_scalar(self):
        self._run_animation(self.to_scalar_group, self.scalar)
    
    def _run_zero(self):
        self._run_animation(self.to_zero_group, 0)
        
    def _run_animation(self, group: QParallelAnimationGroup, end_value):
        for i in range(group.animationCount()):
            anim: QPropertyAnimation = group.animationAt(i)
            target: QGraphicsItem = anim.targetObject()
            anim.setKeyValues([(0, target.scale()), (1, end_value)])
        group.start()
    
    def _on_scalar_ended(self):
        self.state = ConnPathItemsAnimationScaled.State.RUNNING
    
    def _on_zero_ended(self):
        if self.state == ConnPathItemsAnimationScaled.State.TO_ZERO:
            self.con_items_anim.stop()
        self.state = ConnPathItemsAnimationScaled.State.NOT_RUNNING


from qtpy.QtWidgets import (
    QWidget,
    QVBoxLayout,
    QLineEdit,
    QScrollArea,
    QTreeView,
    QSplitter,
    QAbstractItemView,
    QListView,
    QLabel,
)

from qtpy.QtCore import Qt, Signal, QModelIndex, QSortFilterProxyModel
from qtpy.QtGui import QStandardItemModel, QStandardItem, QFont

from ryvencore import Node
from .utils import search, sort_nodes, inc, dec
from ..node_list_widget.NodeWidget import NodeWidget
from statistics import median
from re import escape
from typing import Dict, Type, List

# from ryven import NodesPackage

class NodeStandardItemModel(QStandardItemModel):
    
    def mimeData(self, indexes):
        item = self.itemFromIndex(indexes[0])
        return item.mimeData()
    
class NodeStandardItem(QStandardItem):
    """A node item for use in a model. Helpful when creating a tree view"""
    def __init__(self, node, text = None):
        super().__init__(text)
        self.setDragEnabled(True)
        self.setFont(text_font())
        self.node = node
    
    def mimeData(self):
        return NodeWidget._create_mime_data(self.node) 
    
class NodeListWidget(QWidget):
    # SIGNALS
    escaped = Signal()
    node_chosen = Signal(object)

    def __init__(self, session, show_packages: bool = False):
        super().__init__()

        self.session = session
        self.nodes: List[Type[Node]] = []
        self.package_nodes: List[Type[Node]] = []

        self.current_nodes: List[Node] = []  # currently selectable nodes
        self.active_node_widget_index = -1  # index of focused node widget
        self.active_node_widget = None  # focused node widget
        self.node_widgets: Dict[Node, NodeWidget] = {}  # Node-NodeWidget assignments
        self._node_widget_index_counter = 0

        # holds the path to the tree item
        self.path_to_item: dict = {}
        self.tree_items: List[QStandardItem] = []
        self.show_packages: bool = show_packages
        self._setup_UI()

    def _setup_UI(self) -> None:
        self.main_layout = QVBoxLayout(self)
        self.main_layout.setAlignment(Qt.AlignTop)
        self.setLayout(self.main_layout)

        # splitter between packages and nodes
        splitter = QSplitter(Qt.Vertical)
        self.layout().addWidget(splitter)

        # search for the tree
        self.search_line_tree = QLineEdit(self)
        self.search_line_tree.setPlaceholderText('search packages...')
        self.search_line_tree.textChanged.connect(self.search_pkg_tree)

        # tree view
        self.pack_proxy_model: QSortFilterProxyModel = QSortFilterProxyModel()
        self.pack_proxy_model.setRecursiveFilteringEnabled(True)
        # we need qt6 for not filtering out the children if they would be filtered
        # out otherwise
        self.pack_proxy_model.setFilterCaseSensitivity(Qt.CaseInsensitive)
        self.pack_tree = QTreeView()
        self.pack_tree.setModel(self.pack_proxy_model)
        self.pack_tree.setDragDropMode(QAbstractItemView.DragDropMode.DragOnly)
        self.pack_tree.setDragEnabled(True)

        def on_select(index: QModelIndex):
            source_index = index.model().mapToSource(index)
            item: QStandardItem = index.model().sourceModel().itemFromIndex(source_index)
            func = item.data(Qt.UserRole + 1)
            if func != None:
                func()

        # pkg widget
        self.pkg_widget = QWidget()
        self.pkg_widget.setLayout(QVBoxLayout())
        self.pkg_widget.layout().addWidget(self.search_line_tree)
        self.pkg_widget.layout().addWidget(self.pack_tree)

        self.pack_tree.clicked.connect(on_select)

        if self.show_packages:
            splitter.addWidget(self.pkg_widget)
        
        splitter.setSizes([30])
        
        # nodes widget
        nodes_widget = QWidget()
        nodes_widget.setLayout(QVBoxLayout())
        splitter.addWidget(nodes_widget)

        # adding all stuff to the layout
        self.search_line_edit = QLineEdit(self)
        self.search_line_edit.setPlaceholderText('search for node...')
        self.search_line_edit.textChanged.connect(self._update_view)
        nodes_widget.layout().addWidget(self.search_line_edit)
        
        self.current_pack_label = QLabel('Package: None')
        self.current_pack_label.setFont(text_font())
        nodes_widget.layout().addWidget(self.current_pack_label)
        
        self.list_scroll_area = QScrollArea(self)
        self.list_scroll_area.setVerticalScrollBarPolicy(Qt.ScrollBarAsNeeded)
        self.list_scroll_area.setHorizontalScrollBarPolicy(Qt.ScrollBarAsNeeded)
        self.list_scroll_area.setWidgetResizable(True)
        self.list_scroll_area.setContentsMargins(0, 0, 0, 0)

        self.list_scroll_area_widget = QWidget()
        self.list_scroll_area_widget.setContentsMargins(15, 10, 15, 10)
        self.list_scroll_area.setWidget(self.list_scroll_area_widget)

        self.list_layout = QVBoxLayout()
        self.list_layout.setContentsMargins(0, 0, 0, 0)
        self.list_layout.setAlignment(Qt.AlignTop)
        self.list_scroll_area_widget.setLayout(self.list_layout)

        nodes_widget.layout().addWidget(self.list_scroll_area)

        self._update_view('')

        self.setStyleSheet(self.session.design.node_selection_stylesheet)

        self.search_line_edit.setFocus()

    def search_pkg_tree(self, search: str):
        if search and search != '':
            # removes whitespace and escapes all special regex chars
            new_search = escape(search.strip())
            # regex that enforces the text starts with <new_search>
            self.pack_proxy_model.setFilterRegularExpression(f'^{new_search}')
            self.pack_tree.expandAll()
        else:
            self.pack_proxy_model.setFilterRegularExpression('')
            self.pack_tree.collapseAll()

    def make_nodes_current(self, pack_nodes, pkg_name: str):
        def select_nodes():
            if not pack_nodes or self.package_nodes == pack_nodes:
                return
            self.package_nodes = pack_nodes
            self.current_pack_label.setText(f'Package: {pkg_name}')
            self._update_view()

        return select_nodes

    def make_pack_hier(self) -> None:
        """
        Creates a hierarchical view of the packages based on the nodes' identifier.
        """

        self.tree_items.clear()

        model = NodeStandardItemModel()
        model.setHorizontalHeaderLabels(["Packages"])
        root_item = model.invisibleRootItem()

        # should be dict[str, QStandardItem | (QStandardItem, list)] in 3.9+
        h_dict: dict = {"root_item": root_item}
        font = text_font()
        # A completely nestable tree-view
        for n in self.nodes:
            full_name = n.identifier
            comps = full_name.rsplit('.', 1)
            path = comps[0]
            # node_name = comps[1]

            # if the path isn't found, create all the nested items that are needed
            if not path in h_dict:
                split_path = path.split('.')
                current_path = split_path[0]
                current_root = root_item
                split_path_len = len(split_path)
                for i, s in enumerate(split_path):
                    if not current_path in h_dict:
                        item = QStandardItem(s)
                        item.setFont(font)
                        item.setDragEnabled(False)
                        self.tree_items.append(item)
                        item.setEditable(False)
                        node_list: List[Type[Node]] = []
                        h_dict[current_path] = (item, node_list)
                        item.setData(self.make_nodes_current(node_list, current_path), Qt.UserRole + 1)
                        current_root.appendRow(item)
                    current_root = h_dict[current_path][0]
                    if i != split_path_len - 1:
                        current_path = current_path + f'.{split_path[i+1]}'

            item, pack_nodes = h_dict[path]

            node_item = NodeStandardItem(n, n.title)

            node_item.setEditable(False)
            item.appendRow(node_item)
            pack_nodes.append(n)
            self.tree_items.append(node_item)
        
        self.pack_proxy_model.setSourceModel(model)

    def mousePressEvent(self, event):
        # need to accept the event, so the scene doesn't process it further
        QWidget.mousePressEvent(self, event)
        event.accept()

    def keyPressEvent(self, event):
        """key controls"""

        num_items = len(self.current_nodes)

        if event.key() == Qt.Key_Escape:
            self.escaped.emit()

        elif event.key() == Qt.Key_Down:
            self._set_active_node_widget_index(inc(self.active_node_widget_index, length=num_items))
        elif event.key() == Qt.Key_Up:
            self._set_active_node_widget_index(dec(self.active_node_widget_index, num_items))

        elif event.key() == Qt.Key_Return or event.key() == Qt.Key_Enter:
            if len(self.current_nodes) > 0:
                self._place_node(self.active_node_widget_index)
        else:
            event.setAccepted(False)

    def wheelEvent(self, event):
        # need to accept the event, so the scene doesn't process it further
        QWidget.wheelEvent(self, event)
        event.accept()

    def refocus(self):
        """focuses the search line edit and selects the text"""
        self.search_line_edit.setFocus()
        self.search_line_edit.selectAll()

    def update_list(self, nodes):
        """update the list of available nodes"""
        self.nodes = sort_nodes(nodes)
        self._update_view('')

    def _update_view(self, search_text=''):
        nodes = self.nodes if search_text is not None and search_text != '' else self.package_nodes

        if nodes == None or len(nodes) == 0:
            nodes = self.nodes

        if len(nodes) == 0:
            return

        search_text = search_text.lower()

        # remove all node widgets

        for i in reversed(range(self.list_layout.count())):
            self.list_layout.itemAt(i).widget().setParent(None)

        self.current_nodes.clear()

        self._node_widget_index_counter = 0

        # search
        sorted_distances = search(
            items={n: [n.title.lower()] + n.tags for n in nodes}, text=search_text
        )

        # create node widgets
        cutoff = median(sorted_distances.values())
        for n, dist in sorted_distances.items():
            if search_text != '' and dist > cutoff:
                continue

            self.current_nodes.append(n)

            if self.node_widgets.get(n) is None:
                self.node_widgets[n] = self._create_node_widget(n)

            self.list_layout.addWidget(self.node_widgets[n])

        # focus on first result
        if len(self.current_nodes) > 0:
            self._set_active_node_widget_index(0)

    def _create_node_widget(self, node):
        node_widget = NodeWidget(self, node)
        node_widget.custom_focused_from_inside.connect(self._node_widget_focused_from_inside)
        node_widget.setObjectName('node_widget_' + str(self._node_widget_index_counter))
        self._node_widget_index_counter += 1
        node_widget.chosen.connect(self._node_widget_chosen)

        return node_widget

    def _node_widget_focused_from_inside(self):
        index = self.list_layout.indexOf(self.sender())
        self._set_active_node_widget_index(index)

    def _set_active_node_widget_index(self, index):
        self.active_node_widget_index = index
        node_widget = self.list_layout.itemAt(index).widget()

        if self.active_node_widget:
            self.active_node_widget.set_custom_focus(False)

        node_widget.set_custom_focus(True)
        self.active_node_widget = node_widget
        self.list_scroll_area.ensureWidgetVisible(self.active_node_widget)

    def _node_widget_chosen(self):
        index = int(self.sender().objectName()[self.sender().objectName().rindex('_') + 1 :])
        self._place_node(index)

    def _place_node(self, index):
        node_index = index
        node = self.current_nodes[node_index]
        self.node_chosen.emit(node)
        self.escaped.emit()

def text_font():
    return QFont('Source Code Pro', 9)


# prevent circular imports
from __future__ import annotations
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from ..SessionGUI import SessionGUI

import json

from typing import Tuple, Optional

from qtpy.QtCore import (
    Qt,
    QPointF,
    QPoint,
    QRectF,
    QSizeF,
    Signal,
    QTimer,
    QTimeLine,
    QEvent,
)
from qtpy.QtGui import (
    QPainter,
    QPen,
    QColor,
    QKeySequence,
    QTabletEvent,
    QDropEvent,
    QContextMenuEvent,
    QImage,
    QGuiApplication,
    QFont,
    QTouchEvent,
    QTransform,
)
from qtpy.QtWidgets import (
    QGraphicsView,
    QGraphicsScene,
    QShortcut,
    QMenu,
    QGraphicsItem,
    QPushButton,
    QHBoxLayout,
    QWidget,
)

# for compatibility between qt5 and qt6
try:
    from qtpy.QtGui import QUndoStack
except ImportError:
    from qtpy.QtWidgets import QUndoStack  # type: ignore

from ryvencore.Flow import Flow
from ryvencore.Node import Node
from ryvencore.NodePort import NodePort, NodeInput, NodeOutput

# from ryvencore.Connection import Connection, DataConnection
from ryvencore.InfoMsgs import InfoMsgs
from ryvencore.RC import PortObjPos
from ryvencore.utils import node_from_identifier

from ..GUIBase import GUIBase
from ..utils import *
from .FlowCommands import (
    MoveComponents_Command,
    PlaceNode_Command,
    PlaceDrawing_Command,
    RemoveComponents_Command,
    ConnectPorts_Command,
    Paste_Command,
    FlowUndoCommand,
    SelectComponents_Command,
)
from .FlowViewProxyWidget import *
from .FlowViewStylusModesWidget import FlowViewStylusModesWidget
from .node_list_widget.NodeListWidget import NodeListWidget
from .nodes.NodeGUI import NodeGUI
from .nodes.NodeItem import NodeItem
from .nodes.PortItem import PortItemPin, PortItem
from .connections.ConnectionItem import (
    default_cubic_connection_path,
    ConnectionItem,
    DataConnectionItem,
    ExecConnectionItem,
)
from .drawings.DrawingObject import DrawingObject

from ..Design import Design
from enum import Enum


class _SelectionMode(Enum):
    """
    Determines how selection change events will be handled.
    If INSTANT, then they happen instantly.
    If UNDO, the event is pushed unto the undo stack
    """

    INSTANT = 1
    UNDOABLE_CLICK = 2
    UNDOABLE_RUBBERBANDDRAG = 3


class FlowView(GUIBase, QGraphicsView):
    """Manages the GUI of flows"""

    nodes_selection_changed = Signal(list)
    node_placed = Signal(Node)

    create_node_request = Signal(object, dict)
    remove_node_request = Signal(Node)

    check_connection_validity_request = Signal((NodeOutput, NodeInput), bool)
    connect_request = Signal(NodePort, NodePort)

    # get_nodes_data_request = Signal(list)
    # get_connections_data_request = Signal(list)
    get_flow_data_request = Signal()

    viewport_update_mode_changed = Signal(str)

    def __init__(self, session_gui: SessionGUI, flow: Flow, parent=None) -> None:
        GUIBase.__init__(self, representing_component=flow)
        QGraphicsView.__init__(self, parent=parent)

        # UNDO STACK
        self._undo_stack = QUndoStack(self)
        self._undo_action = self._undo_stack.createUndoAction(self, 'undo')
        self._undo_action.setShortcuts(QKeySequence.Undo)
        self._redo_action = self._undo_stack.createRedoAction(self, 'redo')
        self._redo_action.setShortcuts(QKeySequence.Redo)

        # SHORTCUTS
        self._init_shortcuts()

        # GENERAL ATTRIBUTES
        self.session_gui = session_gui
        self.design: Design = session_gui.design  # type hinting and quicker access

        self.flow: Flow = flow
        self.node_items: Dict = {}  # {Node: NodeItem}
        self.node_items__cache: Dict = {}
        self.connection_items: Dict = {}  # {Connection: ConnectionItem}
        self.connection_items__cache: Dict = {}
        self.selection_mode: _SelectionMode = _SelectionMode.UNDOABLE_CLICK

        # PRIVATE FIELDS
        self._loaded_state: Optional[Dict] = None # h and v scrollbars are changed on import, so we need to defer
        self._tmp_data = None
        self._selected_pin: Optional[PortItemPin] = None
        self._dragging_connection = False
        self._temp_connection_ports: Optional[Tuple[NodeOutput, NodeInput]] = None
        self._waiting_for_connection_request: bool = False
        self.mouse_event_taken = False  # for stylus - see tablet event
        self._last_mouse_move_pos: Optional[QPointF] = None
        self._node_place_pos = QPointF()
        self._left_mouse_pressed_in_flow = False
        self._right_mouse_pressed_in_flow = False
        self._mouse_press_pos: Optional[QPointF] = None
        self._multi_selection = False
        self._current_selected: List[QGraphicsItem] = []
        self._auto_connection_pin: Optional[PortItemPin] = None  # stores the gate that we may try to auto connect to a newly placed NI
        self._panning = False
        self._pan_last_x: Optional[int] = None
        self._pan_last_y: Optional[int] = None
        self._current_scale = 1
        self._total_scale_div = 1
        self._zoom_data = {
            'viewport pos': None,
            'scene pos': None,
            'delta': 0,
        }

        # CONNECTIONS TO FLOW
        self.create_node_request.connect(self.flow.create_node)
        self.remove_node_request.connect(self.flow.remove_node)
        self.check_connection_validity_request.connect(self.flow.check_connection_validity)
        # TODO: need to check if the 2 lines below are used
        # self.get_nodes_data_request.connect(self.flow.gen_nodes_data)
        # self.get_connections_data_request.connect(self.flow.gen_conns_data)
        self.get_flow_data_request.connect(self.flow.data)

        # CONNECTIONS FROM FLOW
        self.flow.node_added.sub(self.add_node)
        self.flow.node_removed.sub(self.remove_node)
        self.flow.connection_added.sub(self.add_connection)
        self.flow.connection_removed.sub(self.remove_connection)
        self.flow.connection_request_valid.sub(self.connection_request_valid)

        # CREATE UI
        scene = QGraphicsScene(self)
        scene.setItemIndexMethod(QGraphicsScene.NoIndex)
        scene.setSceneRect(0, 0, 10000, 7000)

        self.setScene(scene)
        self.setCacheMode(QGraphicsView.CacheBackground)
        self.setViewportUpdateMode(QGraphicsView.BoundingRectViewportUpdate)
        self.setRenderHint(QPainter.Antialiasing)
        self.setTransformationAnchor(QGraphicsView.AnchorUnderMouse)
        self.setDragMode(QGraphicsView.RubberBandDrag)
        self._watch_scene_selection()
        self.setAcceptDrops(True)

        self.centerOn(QPointF(self.viewport().width() / 2, self.viewport().height() / 2))

        self.scene_rect_width = self.mapFromScene(self.sceneRect()).boundingRect().width()
        self.scene_rect_height = self.mapFromScene(self.sceneRect()).boundingRect().height()

        # for proper background painting
        # we could probably use the no cache flag
        on_pan = lambda scroll: self.resetCachedContent()
        
        self.horizontalScrollBar().valueChanged.connect(on_pan)
        self.verticalScrollBar().valueChanged.connect(on_pan)
        
        # NODE LIST WIDGET
        self._node_list_widget = NodeListWidget(self.session_gui)
        self._node_list_widget.setMinimumWidth(260)
        self._node_list_widget.setFixedHeight(300)
        self._node_list_widget.escaped.connect(self.hide_node_list_widget)
        self._node_list_widget.node_chosen.connect(self.create_node__cmd)

        self._node_list_widget_proxy = FlowViewProxyWidget(self)
        self._node_list_widget_proxy.setZValue(1000)
        self._node_list_widget_proxy.setWidget(self._node_list_widget)
        self.scene().addItem(self._node_list_widget_proxy)

        self.hide_node_list_widget()

        # ZOOM WIDGET
        # self._zoom_proxy = FlowViewProxyWidget(self)
        # self._zoom_proxy.setFlag(QGraphicsItem.ItemIgnoresTransformations, True)
        # self._zoom_proxy.setZValue(1001)
        # self._zoom_widget = FlowViewZoomWidget(self)
        # self._zoom_proxy.setWidget(self._zoom_widget)
        # self.scene().addItem(self._zoom_proxy)
        # self.set_zoom_proxy_pos()

        def init_proxy_widget(widget: QWidget, proxy: FlowViewProxyWidget):
            proxy.setFlag(QGraphicsItem.ItemIgnoresTransformations, True)
            proxy.setZValue(1001)
            proxy.setWidget(widget)
            self.scene().addItem(proxy)
            return proxy

        # STYLUS
        self.stylus_mode = ''
        self._current_drawing = None
        self._drawing = False
        self.drawings: List[DrawingObject] = []
        self._stylus_modes_widget = FlowViewStylusModesWidget(self)
        self._stylus_modes_proxy = init_proxy_widget(
            self._stylus_modes_widget, FlowViewProxyWidget(self)
        )
        self.set_stylus_proxy_pos()
        self.setAttribute(Qt.WA_TabletTracking)

        # MENU
        self._menu = QMenu()
        
        # just to add some space for the button
        menu_layout_widget = self._create_no_background_widget("FlowMenu")
        # just to add some space for the button
        menu_layout_widget.setLayout(QHBoxLayout())
        menu_button = QPushButton("Menu")
        self._menu_button = menu_button
        menu_layout_widget.layout().addWidget(menu_button)

        def menu_button_clicked():
            # prob not entirely correct, since menu is part of a layout
            # but since it's the first item, it's the same
            menu_pos = self._menu_button.pos()
            menu_pos = self.mapToGlobal(menu_pos) + QPoint(8, self._menu_button.height() + 10)
            self._menu.exec_(menu_pos)

        menu_button.clicked.connect(menu_button_clicked)

        menu_layout_proxy = init_proxy_widget(menu_layout_widget, FlowViewProxyWidget(self))
        self._menu_widget = menu_button
        self._menu_layout_proxy = menu_layout_proxy
        self.set_menu_proxy_pos()

        # # TOUCH GESTURES
        # recognizer = PanGestureRecognizer()
        # pan_gesture_id = QGestureRecognizer.registerRecognizer(recognizer) <--- CRASH HERE
        # self.grabGesture(pan_gesture_id)
        self.viewport().setAttribute(Qt.WA_AcceptTouchEvents)
        self.last_pinch_points_dist = 0

        # DESIGN
        self.session_gui.design.flow_theme_changed.connect(self._theme_changed)
        self.session_gui.design.performance_mode_changed.connect(self._perf_mode_changed)

        # DATA
        data = self.flow.load_data
        if data is not None:
            view_data = data['flow view']
            if 'drawings' in view_data:  # not all (old) project files have drawings arr
                self.place_drawings_from_data(view_data['drawings'])
            if 'view size' in view_data:
                self.setSceneRect(0, 0, view_data['view size'][0], view_data['view size'][1])

            self._undo_stack.clear()

        # CATCH UP ON FLOW
        for node in self.flow.nodes:
            self.add_node(node)
        for c in [(o, i) for o, conns in self.flow.graph_adj.items() for i in conns]:
            self.add_connection(c)

    def menu(self) -> QMenu:
        """The menu for this flow view"""
        return self._menu

    def _create_no_background_widget(self, name: str):
        result = QWidget()
        result.setObjectName(name)
        result.setStyleSheet(
            f'''
        QWidget#{name} {{
            background: transparent; 
        }}
                '''
        )
        return result

    def _init_shortcuts(self):
        place_new_node_shortcut = QShortcut(QKeySequence('Shift+P'), self)
        place_new_node_shortcut.activated.connect(self._place_new_node_by_shortcut)
        move_selected_components_left_shortcut = QShortcut(QKeySequence('Shift+Left'), self)
        move_selected_components_left_shortcut.activated.connect(self._move_selected_comps_left)
        move_selected_components_up_shortcut = QShortcut(QKeySequence('Shift+Up'), self)
        move_selected_components_up_shortcut.activated.connect(self._move_selected_comps_up)
        move_selected_components_right_shortcut = QShortcut(QKeySequence('Shift+Right'), self)
        move_selected_components_right_shortcut.activated.connect(self._move_selected_comps_right)
        move_selected_components_down_shortcut = QShortcut(QKeySequence('Shift+Down'), self)
        move_selected_components_down_shortcut.activated.connect(self._move_selected_comps_down)
        select_all_shortcut = QShortcut(QKeySequence('Ctrl+A'), self)
        select_all_shortcut.activated.connect(self._select_all_action)
        copy_shortcut = QShortcut(QKeySequence.Copy, self)
        copy_shortcut.activated.connect(self._copy)
        cut_shortcut = QShortcut(QKeySequence.Cut, self)
        cut_shortcut.activated.connect(self._cut)
        paste_shortcut = QShortcut(QKeySequence.Paste, self)
        paste_shortcut.activated.connect(self._paste)

        undo_shortcut = QShortcut(QKeySequence.Undo, self)
        undo_shortcut.activated.connect(self._undo_activated)
        redo_shortcut = QShortcut(QKeySequence.Redo, self)
        redo_shortcut.activated.connect(self._redo_activated)

    def _theme_changed(self, t):
        self._node_list_widget.setStyleSheet(self.session_gui.design.node_selection_stylesheet)
        for n, ni in self.node_items.items():
            ni.widget.rebuild_ui()

        # https://doc.qt.io/qtforpython-5/PySide2/QtWidgets/QGraphicsView.html#PySide2.QtWidgets.PySide2.QtWidgets.QGraphicsView.resetCachedContent
        self.resetCachedContent()
        self.update()

    def _perf_mode_changed(self, mode):
        # set widget value update rule for port inputs, because this takes up quite a bit of performance
        update_widget_value = mode == 'pretty'
        for n, ni in self.node_items.items():
            for inp in ni.inputs:
                inp.update_widget_value = update_widget_value and inp.widget

        self.viewport().update()
        self.scene().update(self.sceneRect())

    def push_undo(self, cmd: FlowUndoCommand):
        self._undo_stack.push(cmd)
        cmd.activate()

    def _undo_activated(self):
        """Triggered by ctrl+z"""
        self._undo_stack.undo()
        self.viewport().update()

    def _redo_activated(self):
        """Triggered by ctrl+y"""
        self._undo_stack.redo()
        self.viewport().update()

    def mousePressEvent(self, event):
        # InfoMsgs.write('mouse press event received, point:', event.pos())

        self._set_selection_mode(_SelectionMode.UNDOABLE_CLICK)

        # to catch tablet events (for some reason, it results in a mousePrEv too)
        if self.mouse_event_taken:
            self.mouse_event_taken = False
            return

        # this allows to chain the selection event when multi-selecting
        hover_item = self.itemAt(event.pos())
        if hover_item is None:
            self._set_selection_mode(_SelectionMode.UNDOABLE_RUBBERBANDDRAG)

        if event.button() != Qt.RightButton:
            QGraphicsView.mousePressEvent(self, event)

        # don't process the event if it has been processed by a specific component in the scene
        # this includes node items, node port pins, proxy widgets etc.
        if self.mouse_event_taken:
            self.mouse_event_taken = False
            self._set_selection_mode(_SelectionMode.UNDOABLE_CLICK)
            return

        if event.button() == Qt.LeftButton:
            if self._node_list_widget_proxy.isVisible():
                self.hide_node_list_widget()

        elif event.button() == Qt.RightButton:
            self._right_mouse_pressed_in_flow = True
            event.accept()

        self._mouse_press_pos = self.mapToScene(event.pos())

    def mouseMoveEvent(self, event):
        if self._right_mouse_pressed_in_flow:  # PAN
            if not self._panning:
                self._panning = True
                self._pan_last_x = event.x()
                self._pan_last_y = event.y()

            self.pan(event.pos())
            event.accept()

        else:
            QGraphicsView.mouseMoveEvent(self, event)

        self._last_mouse_move_pos = self.mapToScene(event.pos())

        if self._dragging_connection:
            self.viewport().repaint()

    def mouseReleaseEvent(self, event):
        # there might be a proxy widget meant to receive the event instead of the flow
        QGraphicsView.mouseReleaseEvent(self, event)

        node_item_at_event_pos = None
        for item in self.items(event.pos()):
            if isinstance(item, NodeItem):
                node_item_at_event_pos = item

        if self.mouse_event_taken:
            self.mouse_event_taken = False
            self.viewport().repaint()
            self._set_selection_mode(_SelectionMode.UNDOABLE_CLICK)
            return

        elif self._panning:
            self._panning = False

        elif event.button() == Qt.RightButton:
            self._right_mouse_pressed_in_flow = False
            if self._mouse_press_pos == self._last_mouse_move_pos and not self.itemAt(event.pos()):
                self.show_place_node_widget(event.pos())
                self._set_selection_mode(_SelectionMode.UNDOABLE_CLICK)
                return

        if self._dragging_connection:
            # connection dropped over port item
            port_items = {i: isinstance(i, PortItem) for i in self.items(event.pos())}
            if any(port_items.values()):
                # gets the first PortItem
                p_i = list(port_items.keys())[list(port_items.values()).index(True)]

                # the validity of the connection is checked in connect_node_ports__cmd
                self.connect_node_ports__cmd(self._selected_pin.port, p_i.port)

            # connection dropped above NodeItem -> auto connect
            elif node_item_at_event_pos:
                # find node item
                ni_under_drop = None
                for item in self.items(event.pos()):
                    if isinstance(item, NodeItem):
                        ni_under_drop = item
                        self.auto_connect(self._selected_pin.port, ni_under_drop.node)
                        break

            # connection dropped somewhere else -> show node choice widget
            else:
                self._auto_connection_pin = self._selected_pin
                self.show_place_node_widget(event.pos())

            self._dragging_connection = False
            self._selected_pin = None

        # if event.button() == Qt.LeftButton:
        #     self._left_mouse_pressed_in_flow = False
        elif event.button() != Qt.RightButton:
            if self.selection_mode == _SelectionMode.UNDOABLE_RUBBERBANDDRAG:
                self._select__cmd()

        else:
            self._right_mouse_pressed_in_flow = False

        self._set_selection_mode(_SelectionMode.UNDOABLE_CLICK)
        self.viewport().repaint()

    def keyPressEvent(self, event):
        QGraphicsView.keyPressEvent(self, event)
        if event.isAccepted():
            return

        if event.key() == Qt.Key_Escape:  # do I need that... ?
            self.clearFocus()
            self.setFocus()
            return True

        elif event.key() == Qt.Key_Delete and self.selection_mode != _SelectionMode.UNDOABLE_RUBBERBANDDRAG:
            self.remove_selected_components__cmd()

    def wheelEvent(self, event):
        if event.modifiers() & Qt.ControlModifier:
            event.accept()

            view_pos = event.position()
            self._zoom_data['viewport pos'] = view_pos
            self._zoom_data['scene pos'] = self.mapToScene(view_pos.toPoint())

            y_delta = event.angleDelta().y()
            self._zoom_data['delta'] += y_delta

            if self._zoom_data['delta'] * y_delta < 0:
                self._zoom_data['delta'] = y_delta

            anim = QTimeLine(100, self)
            anim.setUpdateInterval(10)
            anim.valueChanged.connect(self._scaling_time)
            anim.start()

        else:
            super().wheelEvent(event)

    def _scaling_time(self, x):
        delta = self._zoom_data['delta'] / 8
        if abs(delta) <= 5:
            delta = self._zoom_data['delta']
        self._zoom_data['delta'] -= delta

        self.zoom(self._zoom_data['viewport pos'], self._zoom_data['scene pos'], delta)

    def viewportEvent(self, event: QEvent) -> bool:
        """handling some touch features here"""

        if event.type() == QEvent.TouchBegin:
            self.setDragMode(QGraphicsView.NoDrag)
            return True

        elif event.type() == QEvent.TouchUpdate:
            if len(event.touchPoints()) == 2:
                tp0, tp1 = event.touchPoints()[0], event.touchPoints()[1]

                p0, p1 = tp0.pos(), tp1.pos()
                # sp0, sp1 = tp0.scenePos(), tp1.scenePos()

                pinch_points_dist = points_dist(p0, p1)

                if self.last_pinch_points_dist == 0:
                    self.last_pinch_points_dist = pinch_points_dist

                center = middle_point(p0, p1)
                self.zoom(
                    p_abs=center,
                    p_mapped=self.mapToScene(center.toPoint()),
                    angle=((pinch_points_dist / self.last_pinch_points_dist) ** 10 - 1) * 100,
                )

                self.last_pinch_points_dist = pinch_points_dist

            return True

        elif event.type() == QEvent.TouchEnd:
            self.last_pinch_points_dist = 0
            self.setDragMode(QGraphicsView.RubberBandDrag)
            return True

        else:
            return super().viewportEvent(event)

    def tabletEvent(self, event: QTabletEvent) -> None:
        """tabletEvent gets called by stylus operations.
        LeftButton: std, no button pressed
        RightButton: upper button pressed"""

        # if in edit mode and not panning or starting a pan, pass on to std mouseEvent handlers above
        if (
            self.stylus_mode == 'edit'
            and not self._panning
            and not (event.type() == QTabletEvent.TabletPress and event.button() == Qt.RightButton)
        ):
            return  # let the mousePress/Move/Release-Events handle it

        scaled_event_pos: QPointF = event.posF()
        scaled_event_pos /= self._current_scale  # type: ignore
        # mypy thinks the above results in a float, but it doesn't

        if event.type() == QTabletEvent.TabletPress:
            self.mouse_event_taken = True

            if event.button() == Qt.LeftButton:
                if self.stylus_mode == 'comment':
                    view_pos = self.mapToScene(self.viewport().pos())
                    new_drawing = self._create_and_place_drawing__cmd(
                        view_pos + scaled_event_pos,
                        data={
                            **self._stylus_modes_widget.get_pen_settings(),
                            'viewport pos': view_pos,
                        },
                    )
                    self._current_drawing = new_drawing
                    self._drawing = True
            elif event.button() == Qt.RightButton:
                self._panning = True
                self._pan_last_x = event.x()
                self._pan_last_y = event.y()

        elif event.type() == QTabletEvent.TabletMove:
            self.mouse_event_taken = True
            if self._panning:
                self.pan(event.pos())

            elif event.pointerType() == QTabletEvent.Eraser:
                if self.stylus_mode == 'comment':
                    for i in self.items(event.pos()):
                        if isinstance(i, DrawingObject):
                            self.remove_drawing(i)
                            break
            elif self.stylus_mode == 'comment' and self._drawing:
                assert self._current_drawing is not None
                if self._current_drawing.append_point(scaled_event_pos):
                    self._current_drawing.stroke_weights.append(
                        event.pressure() * self._stylus_modes_widget.pen_width()
                    )
                self._current_drawing.update()
                self.viewport().update()

        elif event.type() == QTabletEvent.TabletRelease:
            if self._panning:
                self._panning = False
            if self.stylus_mode == 'comment' and self._drawing:
                assert self._current_drawing is not None
                self._current_drawing.finish()
                InfoMsgs.write('drawing finished')
                self._current_drawing = None
                self._drawing = False

    """
    --> https://forum.qt.io/topic/121473/qgesturerecognizer-registerrecognizer-crashes-using-pyside2

    def event(self, event) -> bool:
        # if event.type() == QEvent.Gesture:
        #     if event.gesture(PanGesture) is not None:
        #         return self.pan_gesture(event)

        return QGraphicsView.event(self, event)

    def pan_gesture(self, event: QGestureEvent) -> bool:
        pan: PanGesture = event.gesture(PanGesture)
        print(pan)
        return True
    """

    def dragEnterEvent(self, event):
        if event.mimeData().hasFormat('application/json'):
            event.acceptProposedAction()

    def dragMoveEvent(self, event):
        if event.mimeData().hasFormat('application/json'):
            event.acceptProposedAction()

    def dropEvent(self, event: QDropEvent):
        try:
            text = str(event.mimeData().data('application/json'), 'utf-8')
            data: Dict = json.loads(text)

            if data['type'] == 'node':
                self._node_place_pos = self.mapToScene(event.pos())
                self.create_node__cmd(
                    node_from_identifier(
                        data['node identifier'],
                        list(self.session_gui.core_session.nodes)
                    )
                )
            # without this keyPressed function isn't called if we don't click in the view
            self.setFocus()
        except Exception:
            pass
    
    def contextMenuEvent(self, event: QContextMenuEvent):
        QGraphicsView.contextMenuEvent(self, event)
        # in the case of the menu already being shown by a widget under the mouse, the event is accepted here
        if event.isAccepted():
            return
        
        for i in self.items(event.pos()):
            if isinstance(i, NodeItem):
                ni: NodeItem = i
                menu: QMenu = ni.get_context_menu()
                menu.exec_(event.globalPos())
                event.accept()

    # PAINTING
    def drawBackground(self, painter, rect):
        painter.setBrush(self.session_gui.design.flow_theme.flow_background_brush)
        painter.setPen(Qt.NoPen)
        painter.drawRect(self.sceneRect())

        if self.session_gui.design.performance_mode == 'pretty':
            theme = self.session_gui.design.flow_theme
            if theme.flow_background_grid and self._current_scale >= 0.7:
                if theme.flow_background_grid[0] == 'points':
                    color = theme.flow_background_grid[1]
                    pen_width = theme.flow_background_grid[2]
                    diff_x = theme.flow_background_grid[3]
                    diff_y = theme.flow_background_grid[4]

                    pen = QPen(color)
                    pen.setWidthF(pen_width)
                    painter.setPen(pen)

                    points_rect = rect.toRect()
                    top_left = points_rect.topLeft()
                    for x in range(
                        diff_x + top_left.x(), points_rect.width() + top_left.x(), diff_x
                    ):
                        for y in range(
                            diff_y + top_left.y(), points_rect.height() + top_left.y(), diff_y
                        ):
                            painter.drawPoint(x, y)

        # has to be called here instead of in drawForeground to prevent lagging
        self.set_stylus_proxy_pos()
        self.set_menu_proxy_pos()
        # self.set_zoom_proxy_pos()

    def drawForeground(self, painter, rect):
        # DRAW CURRENTLY DRAGGED CONNECTION
        if self._dragging_connection:
            pen = QPen(QColor('#101520'))
            pen.setWidth(3)
            pen.setStyle(Qt.DotLine)
            painter.setPen(pen)

            pin_pos = self._selected_pin.get_scene_center_pos()
            spp = self._selected_pin.port
            cursor_pos = self._last_mouse_move_pos

            pos1 = pin_pos if spp.io_pos == PortObjPos.OUTPUT else cursor_pos
            pos2 = pin_pos if spp.io_pos == PortObjPos.INPUT else cursor_pos

            painter.drawPath(default_cubic_connection_path(pos1, pos2))

        # # DRAW SELECTED NIs BORDER
        # for ni in self.selected_node_items():
        #     pen = QPen(self.session.design.flow_theme.flow_highlight_pen_color)
        #     pen.setWidth(3)
        #     painter.setPen(pen)
        #     painter.setBrush(Qt.NoBrush)
        #
        #     size_factor = 1.2
        #     x = ni.pos().x() - ni.boundingRect().width() / 2 * size_factor
        #     y = ni.pos().y() - ni.boundingRect().height() / 2 * size_factor
        #     w = ni.boundingRect().width() * size_factor
        #     h = ni.boundingRect().height() * size_factor
        #     painter.drawRoundedRect(x, y, w, h, 10, 10)

        # DRAW SELECTED DRAWINGS BORDER
        for p_o in self.selected_drawings():
            pen = QPen(QColor('#a3cc3b'))
            pen.setWidth(2)
            painter.setPen(pen)
            painter.setBrush(Qt.NoBrush)

            size_factor = 1.05
            x = p_o.pos().x() - p_o.width / 2 * size_factor
            y = p_o.pos().y() - p_o.height / 2 * size_factor
            w = p_o.width * size_factor
            h = p_o.height * size_factor
            painter.drawRoundedRect(x, y, w, h, 6, 6)
            painter.drawEllipse(p_o.pos().x(), p_o.pos().y(), 2, 2)

    def get_viewport_img(self) -> QImage:
        """Returns a clear image of the viewport"""

        self.hide_proxies()
        img = QImage(
            size=QSizeF(
                self.viewport().rect().width(),
                self.viewport().height(),
            ),
            format=QImage.Format_ARGB32,
        )
        img.fill(Qt.transparent)

        painter = QPainter(img)
        painter.setRenderHint(QPainter.Antialiasing)
        self.render(painter, self.viewport().rect(), self.viewport().rect())
        self.show_proxies()
        return img

    def get_whole_scene_img(self) -> QImage:
        """Returns an image of the whole scene, scaled accordingly to current scale factor.
        Due to a bug this only works from the viewport position down and right, so the user has to scroll to
        the top left corner in order to get the full scene"""

        self.hide_proxies()
        img = QImage(
            size=QSizeF(
                self.sceneRect().width() / self._total_scale_div,
                self.sceneRect().height() / self._total_scale_div,
            ),
            format=QImage.Format_RGB32,
        )
        img.fill(Qt.transparent)

        painter = QPainter(img)
        painter.setRenderHint(QPainter.Antialiasing)
        rect = QRectF()
        rect.setLeft(-self.viewport().pos().x())
        rect.setTop(-self.viewport().pos().y())
        rect.setWidth(img.rect().width())
        rect.setHeight(img.rect().height())
        # rect is right... but it only renders from the viewport's point down-and rightwards, not from topleft (0,0) ...
        self.render(painter, rect, rect.toRect())
        self.show_proxies()
        return img

    # PROXY POSITIONS

    # def set_zoom_proxy_pos(self):
    #     self._zoom_proxy.setPos(self.mapToScene(self.viewport().width() - self._zoom_widget.width(), 0))

    def set_stylus_proxy_pos(self):
        self._stylus_modes_proxy.setPos(
            self.mapToScene(self.viewport().width() - self._stylus_modes_widget.width(), 0)
        )
        # self.mapToScene(self.viewport().width() - self._stylus_modes_widget.width() - self._zoom_widget.width(), 0))
    
    def set_menu_proxy_pos(self):
        self._menu_layout_proxy.setPos(self.mapToScene(0, 0))
        
    def hide_proxies(self):
        self._stylus_modes_proxy.hide()
        self._menu_layout_proxy.hide()
        # self._zoom_proxy.hide()

    def show_proxies(self):
        self._stylus_modes_proxy.show()
        self._menu_layout_proxy.show()
        # self._zoom_proxy.show()

    # PLACE NODE WIDGET
    def show_place_node_widget(self, pos, nodes=None):
        """Opens the place node dialog in the scene."""

        # calculating position
        self._node_place_pos = self.mapToScene(pos)
        dialog_pos = QPoint(pos.x() + 1, pos.y() + 1)

        # ensure that the node_list_widget stays in the viewport
        if (
            dialog_pos.x() + self._node_list_widget.width() / self._total_scale_div
            > self.viewport().width()
        ):
            dialog_pos.setX(
                dialog_pos.x()
                - (
                    dialog_pos.x()
                    + self._node_list_widget.width() / self._total_scale_div
                    - self.viewport().width()
                )
            )
        if (
            dialog_pos.y() + self._node_list_widget.height() / self._total_scale_div
            > self.viewport().height()
        ):
            dialog_pos.setY(
                dialog_pos.y()
                - (
                    dialog_pos.y()
                    + self._node_list_widget.height() / self._total_scale_div
                    - self.viewport().height()
                )
            )
        dialog_pos = self.mapToScene(dialog_pos)

        # open nodes dialog
        # the dialog emits 'node_chosen' which is connected to self.place_node__cmd
        self._node_list_widget.update_list(
            nodes if nodes is not None else self.session_gui.core_session.nodes
        )
        self._node_list_widget_proxy.setPos(dialog_pos)
        self._node_list_widget_proxy.show()
        self._node_list_widget.refocus()

    def hide_node_list_widget(self):
        self._node_list_widget_proxy.hide()
        self._node_list_widget.clearFocus()
        self._auto_connection_pin = None

    def _place_new_node_by_shortcut(self):  # Shift+P
        point_in_viewport = None
        selected_NIs = self.selected_node_items()
        if len(selected_NIs) > 0:
            x = selected_NIs[-1].pos().x() + 150
            y = selected_NIs[-1].pos().y()
            self._node_place_pos = QPointF(x, y)
            point_in_viewport = self.mapFromScene(QPoint(x, y))
        else:  # place in center
            viewport_x = self.viewport().width() / 2
            viewport_y = self.viewport().height() / 2
            point_in_viewport = QPointF(viewport_x, viewport_y).toPoint()
            self._node_place_pos = self.mapToScene(point_in_viewport)

        self.show_place_node_widget(point_in_viewport)

    # PAN
    def pan(self, new_pos):
        self.horizontalScrollBar().setValue(
            self.horizontalScrollBar().value() - (new_pos.x() - self._pan_last_x)
        )
        self.verticalScrollBar().setValue(
            self.verticalScrollBar().value() - (new_pos.y() - self._pan_last_y)
        )
        self._pan_last_x = new_pos.x()
        self._pan_last_y = new_pos.y()

    # ZOOM
    def zoom_in(self, amount):
        local_viewport_center = QPoint(self.viewport().width() / 2, self.viewport().height() / 2)
        self.zoom(local_viewport_center, self.mapToScene(local_viewport_center), amount)

    def zoom_out(self, amount):
        local_viewport_center = QPoint(self.viewport().width() / 2, self.viewport().height() / 2)
        self.zoom(local_viewport_center, self.mapToScene(local_viewport_center), -amount)

    def zoom(self, p_abs, p_mapped, angle):
        # print(f'zooming to {p_abs} / {p_mapped} with angle: {angle}')

        by = 0
        velocity = 2 * (1 / self._current_scale) + 0.5
        if velocity > 3:
            velocity = 3
        if self._current_scale < 1:
            velocity *= self._current_scale

        zoom_dir_IN = angle > 0
        if zoom_dir_IN:
            by = 1 + (angle / 4000) * velocity
        else:
            by = 1 - (-angle / 4000) * velocity

        if zoom_dir_IN:
            if self._current_scale * by < 3:
                self.scale(by, by)
                self._current_scale *= by
        else:
            if (
                self.scene_rect_width * by >= self.viewport().size().width()
                and self.scene_rect_height * by >= self.viewport().size().height()
            ):
                self.scale(by, by)
                self._current_scale *= by

        w = self.viewport().width()
        h = self.viewport().height()
        wf = self.mapToScene(QPoint(w - 1, 0)).x() - self.mapToScene(QPoint(0, 0)).x()
        hf = self.mapToScene(QPoint(0, h - 1)).y() - self.mapToScene(QPoint(0, 0)).y()
        lf = p_mapped.x() - p_abs.x() * wf / w
        tf = p_mapped.y() - p_abs.y() * hf / h

        self.ensureVisible(lf, tf, wf, hf, 0, 0)

        target_rect = QRectF(QPointF(lf, tf), QSizeF(wf, hf))
        self._total_scale_div = target_rect.width() / self.viewport().width()

        self.ensureVisible(target_rect, 0, 0)

    # NODES
    def create_node__cmd(self, node_class):
        self.push_undo(PlaceNode_Command(self, node_class, self._node_place_pos))

    def add_node(self, node: Node):
        # create item
        item: NodeItem

        if node in self.node_items__cache.keys():  # load from cache
            # print('using a cached item')
            item = self.node_items__cache[node]
            self._add_node_item(item)

        else:  # create new item
            item = NodeItem(
                node=node,
                node_gui=
                    (node.GUI if hasattr(node, 'GUI') else NodeGUI)     # use custom GUI class if available
                    ((node, self.session_gui)),                         # calls __init__ of NodeGUI class with tuple arg
                flow_view=self,
                design=self.session_gui.design,
            )
            item.initialize()

            self.node_placed.emit(node)

            item_data = node.load_data
            if item_data is not None and 'pos x' in item_data:
                pos = QPointF(item_data['pos x'], item_data['pos y'])
            else:
                pos = self._node_place_pos

            self._add_node_item(item, pos)

        # auto connect
        if self._auto_connection_pin:
            self.auto_connect(self._auto_connection_pin.port, node)

    def _add_node_item(self, item: NodeItem, pos=None):
        self.node_items[item.node] = item

        self._set_selection_mode(_SelectionMode.INSTANT)

        self.scene().addItem(item)
        if pos:
            item.setPos(pos)

        # select new item
        self.clear_selection()
        item.setSelected(True)

    def remove_node(self, node):
        item = self.node_items[node]
        self._remove_node_item(item)
        del self.node_items[node]

    def _remove_node_item(self, item: NodeItem):
        # store item in case the remove action gets undone later
        self._set_selection_mode(_SelectionMode.INSTANT)
        self.node_items__cache[item.node] = item
        self.scene().removeItem(item)

    # CONNECTIONS
    def connect_node_ports__cmd(self, p1: NodePort, p2: NodePort):
        # Need to check order of ports since Flow.check_connection_validity needs (NodeOutput, NodeInput)
        if isinstance(p1, NodeOutput) and isinstance(p2, NodeInput):
            self._temp_connection_ports = (p1, p2)
            self._waiting_for_connection_request = True
            self.check_connection_validity_request.emit((p1, p2), True)

        elif isinstance(p1, NodeInput) and isinstance(p2, NodeOutput):
            self._temp_connection_ports = (p2, p1)
            self._waiting_for_connection_request = True
            self.check_connection_validity_request.emit((p2, p1), True)

        else:
            self.connection_request_valid(False)

    def connection_request_valid(self, valid: bool):
        """
        Triggered after the abstract flow evaluated validity of pending connect request.
        This can also lead to a disconnect!
        """
        
        # TODO: this stuff is too complicated, simplify

        if self._waiting_for_connection_request:
            self._waiting_for_connection_request = False
        else:
            return

        assert self._temp_connection_ports is not None

        if valid:
            out: NodeOutput
            inp: NodeInput
            out, inp = self._temp_connection_ports

            if self.flow.graph_adj_rev[inp] not in (None, out): # out connected to something else
                # remove existing connection
                self.push_undo(
                    ConnectPorts_Command(self, out=self.flow.graph_adj_rev[inp], inp=inp)
                )

            if self.flow.connected_output(inp) == out:
                # if the exact connection exists, we want to remove it by command
                self.push_undo(
                    ConnectPorts_Command(self, out=self.flow.connected_output(inp), inp=inp)
                )
            else:
                self.push_undo(ConnectPorts_Command(self, out=out, inp=inp))

    def add_connection(self, c: Tuple[NodeOutput, NodeInput]):
        out, inp = c

        # TODO: need to verify that connection_items_cache still works fine with new connection object
        item: ConnectionItem
        if c in self.connection_items__cache.keys():
            item = self.connection_items__cache[c]
        else:
            if inp.type_ == 'data':
                # item = self.CLASSES['data conn item'](c, self.session.design)
                item = DataConnectionItem(c, self.session_gui.design)
            else:
                # item = self.CLASSES['exec conn item'](c, self.session.design)
                item = ExecConnectionItem(c, self.session_gui.design)

        self._add_connection_item(item)

        item.out_item.port_connected()
        item.inp_item.port_connected()
        # for updating the visual state
        for i in c:
            self.node_items[i.node].update()

    def _add_connection_item(self, item: ConnectionItem):
        self._set_selection_mode(_SelectionMode.INSTANT)
        self.connection_items[item.connection] = item
        self.scene().addItem(item)
        item.recompute()
        item.setZValue(-1)
        # self.viewport().repaint()

    def remove_connection(self, c: Tuple[NodeOutput, NodeInput]):
        item = self.connection_items[c]
        self._remove_connection_item(item)

        item.out_item.port_disconnected()
        item.inp_item.port_disconnected()
        # for updating the visual state
        for i in c:
            self.node_items[i.node].update()
        del self.connection_items[c]

    def _remove_connection_item(self, item: ConnectionItem):
        self._set_selection_mode(_SelectionMode.INSTANT)
        self.connection_items__cache[item.connection] = item
        self.scene().removeItem(item)

    def auto_connect(self, p: NodePort, n: Node):
        if p.io_pos == PortObjPos.OUTPUT:
            for inp in n.inputs:
                if p.type_ == inp.type_:
                    # connect exactly once
                    self.connect_node_ports__cmd(p, inp)
                    return
        elif p.io_pos == PortObjPos.INPUT:
            for out in n.outputs:
                if p.type_ == out.type_:
                    # connect exactly once
                    self.connect_node_ports__cmd(p, out)
                    return

    def update_conn_item(self, c: Tuple[NodeOutput, NodeInput]):
        if c in self.connection_items:
            self.connection_items[c].changed = True
            self.connection_items[c].update()

    # DRAWINGS
    def create_drawing(self, data=None) -> DrawingObject:
        """Creates and returns a new DrawingObject."""

        new_drawing = DrawingObject(self, data)
        return new_drawing

    def add_drawing(self, drawing_obj: DrawingObject, posF=None):
        """Adds a DrawingObject to the scene."""

        self._set_selection_mode(_SelectionMode.INSTANT)
        self.scene().addItem(drawing_obj)
        if posF:
            drawing_obj.setPos(posF)
        self.drawings.append(drawing_obj)

    def add_drawings(self, drawings):
        """Adds a list of DrawingObjects to the scene."""

        for d in drawings:
            self.add_drawing(d)

    def remove_drawing(self, drawing: DrawingObject):
        """Removes a drawing from the scene."""

        # TODO https://github.com/leon-thomm/ryvencore/issues/4
        self._set_selection_mode(_SelectionMode.INSTANT)
        self.scene().removeItem(drawing)
        self.drawings.remove(drawing)

    def place_drawings_from_data(self, drawings_data: list, offset_pos=QPoint(0, 0)):
        """Creates and places drawings from drawings. The same list is returned by the data_() method
        at 'drawings'."""

        new_drawings = []
        for d_data in drawings_data:
            x = d_data['pos x'] + offset_pos.x()
            y = d_data['pos y'] + offset_pos.y()
            new_drawing = self.create_drawing(data=d_data)
            self.add_drawing(new_drawing, QPointF(x, y))
            new_drawings.append(new_drawing)

        return new_drawings

    def _create_and_place_drawing__cmd(self, posF, data=None):
        new_drawing_obj = self.create_drawing(data)
        place_command = PlaceDrawing_Command(self, posF, new_drawing_obj)
        self.push_undo(place_command)
        return new_drawing_obj

    # ADDING/REMOVING COMPONENTS
    def add_component(self, e: QGraphicsItem):
        if isinstance(e, NodeItem):
            self.add_node(e.node)
        elif isinstance(e, DrawingObject):
            self.add_drawing(e)

    def remove_components(self, comps: List[QGraphicsItem]):
        for c in comps:
            self.remove_component(c)

    def remove_component(self, e: QGraphicsItem):
        if isinstance(e, NodeItem):
            self.remove_node(e.node)
        elif isinstance(e, DrawingObject):
            self.remove_drawing(e)

    def remove_selected_components__cmd(self):
        if self._current_selected:
            self.push_undo(RemoveComponents_Command(self, self._current_selected))

        self.viewport().update()

    # MOVING COMPONENTS
    def _move_selected_copmonents__cmd(self, pos_diff: QPointF, already_moved=False):

        # if one node item would leave the scene (f.ex. pos.x < 0), stop
        left = False
        for i in self.scene().selectedItems():
            new_pos = i.pos() + pos_diff
            w = i.boundingRect().width()
            h = i.boundingRect().height()
            if (
                new_pos.x() - w / 2 < 0
                or new_pos.x() + w / 2 > self.scene().width()
                or new_pos.y() - h / 2 < 0
                or new_pos.y() + h / 2 > self.scene().height()
            ):
                left = True
                break

        if not left:
            # moving the items
            if not already_moved:
                items_group = self.scene().createItemGroup(self.scene().selectedItems())
                items_group.moveBy(pos_diff.x(), pos_diff.y())
                self.scene().destroyItemGroup(items_group)

            # saving the command
            self.push_undo(
                MoveComponents_Command(
                    self,
                    self.scene().selectedItems(),
                    p_from=-pos_diff,
                    p_to=QPointF(0, 0),
                )
            )

        self.viewport().repaint()

    def _move_selected_comps_left(self):
        self._move_selected_copmonents__cmd(QPointF(-40, 0))

    def _move_selected_comps_up(self):
        self._move_selected_copmonents__cmd(QPointF(0, -40))

    def _move_selected_comps_right(self):
        self._move_selected_copmonents__cmd(QPointF(+40, 0))

    def _move_selected_comps_down(self):
        self._move_selected_copmonents__cmd(QPointF(0, +40))

    # SELECTION
    def _scene_selection_changed(self):
        if self.selection_mode == _SelectionMode.INSTANT:
            self._current_selected = self.scene().selectedItems()
            self.emit_selected_items()
        elif self.selection_mode == _SelectionMode.UNDOABLE_CLICK:
            self._select__cmd()

    def _watch_scene_selection(self):
        self.scene().selectionChanged.connect(self._scene_selection_changed)

    def _unwatch_scene_selection(self):
        self.scene().selectionChanged.disconnect(self._scene_selection_changed)

    def emit_selected_items(self):
        """Emits signals indicating selection changes."""

        self.nodes_selection_changed.emit(self.selected_nodes())

    def _set_selection_mode(self, mode: _SelectionMode):
        self.selection_mode = mode

    def _select__cmd(self):
        items = self.scene().selectedItems()
        if items != self._current_selected:
            self.push_undo(
                SelectComponents_Command(self, items, self._current_selected)
            )

    def selected_node_items(self, item_list: Optional[List[NodeItem]] = None) -> List[NodeItem]:
        """Returns a list of the currently selected NodeItems."""

        search_list = item_list if item_list is not None else self.scene().selectedItems()
        return [node_item for node_item in search_list if isinstance(node_item, NodeItem)]

    def selected_nodes(self, item_list: Optional[List[NodeItem]] = None) -> List[Node]:
        """Returns a list of the currently selected nodes."""
        
        return [node_item.node for node_item in self.selected_node_items(item_list)]

    def selected_drawings(self) -> List[DrawingObject]:
        """Returns a list of the currently selected drawings."""

        return [
            drawing
            for drawing in self.scene().selectedItems()
            if isinstance(drawing, DrawingObject)
        ]

    def select_items(self, items: list):
        """
        Selects a list of items in the scene.
        It always clears the current selection
        Does not cause any command to be pushed to the undo stack.
        Emits signals indicating selection changes.
        """
        
        self._unwatch_scene_selection()
        self._current_selected = items
        for i in self.scene().items():
            if i.flags() & QGraphicsItem.ItemIsSelectable != 0:
                i.setSelected(False)
                
        for i in items:
            if i.flags() & QGraphicsItem.ItemIsSelectable != 0:
                i.setSelected(True)
        self.emit_selected_items()
        self.viewport().repaint()
        self._watch_scene_selection()

    def select_all(self):
        self.select_items(self.scene().items())

    def clear_selection(self):
        self.select_items([])

    # ACTIONS
    def _select_all_action(self):  # ctrl+a
        all_items = self.scene().items()
        if all_items != self._current_selected:
            self.push_undo(SelectComponents_Command(self, all_items, self._current_selected))

    def _copy(self):  # ctrl+c
        data = {
            'nodes': self._get_nodes_data(self.selected_nodes()),
            'connections': self._get_connections_data(self.selected_nodes()),
            'output data': self._get_output_data(self.selected_nodes()),
            'drawings': self._get_drawings_data(self.selected_drawings()),
        }
        QGuiApplication.clipboard().setText(json.dumps(data))

    def _cut(self):  # ctrl+x
        data = {
            'nodes': self._get_nodes_data(self.selected_nodes()),
            'connections': self._get_connections_data(self.selected_nodes()),
            'drawings': self._get_drawings_data(self.selected_drawings()),
        }
        QGuiApplication.clipboard().setText(json.dumps(data))
        self.remove_selected_components__cmd()

    def _paste(self):  # ctrl+v
        data = {}
        try:
            data = json.loads(QGuiApplication.clipboard().text())
        except Exception as e:
            return

        self.clear_selection()

        # calculate offset
        positions = []
        for d in data['drawings']:
            positions.append({
                'x': d['pos x'], 
                'y': d['pos y']
            })
        for n in data['nodes']:
            positions.append({
                'x': n['pos x'], 
                'y': n['pos y']
            })

        offset_for_middle_pos = QPointF(0, 0)
        if len(positions) > 0:
            rect = QRectF(positions[0]['x'], positions[0]['y'], 0, 0)
            for p in positions:
                x = p['x']
                y = p['y']
                if x < rect.left():
                    rect.setLeft(x)
                if x > rect.right():
                    rect.setRight(x)
                if y < rect.top():
                    rect.setTop(y)
                if y > rect.bottom():
                    rect.setBottom(y)

            offset_for_middle_pos = self._last_mouse_move_pos - rect.center()

        self.push_undo(Paste_Command(self, data, offset_for_middle_pos))

    # DATA
    def complete_data(self, data: Dict):
        data['flow view'] = {
            'drawings': self._get_drawings_data(self.drawings),
            'view size': [
                self.sceneRect().size().width(),
                self.sceneRect().size().height(),
            ],
        }

        return data

    def _get_nodes_data(self, nodes):
        """generates the data for the specified list of nodes"""
        f_complete_data = self.session_gui.core_session.complete_data
        return f_complete_data(self.flow._gen_nodes_data(nodes))

    def _get_connections_data(self, nodes):
        """generates the connections data for connections between a specified list of nodes"""
        f_complete_data = self.session_gui.core_session.complete_data
        return f_complete_data(self.flow._gen_conns_data(nodes))

    def _get_output_data(self, nodes):
        """generates the serialized data of output ports of the specified nodes"""
        f_complete_data = self.session_gui.core_session.complete_data
        return f_complete_data(self.flow._gen_output_data(nodes))

    def _get_drawings_data(self, drawings):
        """generates the data for a list of drawings"""

        return [d.data() for d in drawings]

    def save_state(self) -> dict:
        """Save the state of the view in a json-compliant dict"""
        transform = self.transform()
        return {
            'm11': transform.m11(),  # Horizontal scaling
            'm12': transform.m12(),  # Horizontal shearing
            'm13': transform.m13(),  # Horizontal projection
            'm21': transform.m21(),  # Vertical shearing
            'm22': transform.m22(),  # Vertical scaling
            'm23': transform.m23(),  # Vertical projection
            'm31': transform.m31(),  # Horizontal translation
            'm32': transform.m32(),  # Vertical translation
            'm33': transform.m33(),  # Perspective division
            'v_scroll': self.horizontalScrollBar().value(),
            'h_scroll': self.verticalScrollBar().value(),
        }
    
    def load(self, state: Dict):
        """Load the state of the view"""
        transform = QTransform(
            state['m11'], state['m12'], state['m13'],
            state['m21'], state['m22'], state['m23'],
            state['m31'], state['m32'], state['m33']
        )
        
        self.setTransform(transform)
        self.horizontalScrollBar().setValue(state['v_scroll'])
        self.verticalScrollBar().setValue(state['h_scroll'])
        
        self._loaded_state = state
    
    def reload(self):
        if self._loaded_state is not None:
            self.load(self._loaded_state)
        


# prevent circular imports
from __future__ import annotations
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from .FlowView import FlowView

from qtpy.QtWidgets import QGraphicsProxyWidget, QGraphicsSceneHoverEvent


class FlowViewProxyWidget(QGraphicsProxyWidget):
    """Ensures easy controls event handling for QProxyWidgets in the flow."""

    def __init__(self, flow_view, parent=None):
        super(FlowViewProxyWidget, self).__init__(parent)

        self.flow_view = flow_view

    def mousePressEvent(self, arg__1):
        QGraphicsProxyWidget.mousePressEvent(self, arg__1)
        if arg__1.isAccepted():
            self.flow_view.mouse_event_taken = True

    def mouseReleaseEvent(self, arg__1):
        self.flow_view.mouse_event_taken = True
        QGraphicsProxyWidget.mouseReleaseEvent(self, arg__1)

    def wheelEvent(self, event):
        QGraphicsProxyWidget.wheelEvent(self, event)

    def keyPressEvent(self, arg__1):
        QGraphicsProxyWidget.keyPressEvent(self, arg__1)
        if arg__1.isAccepted():
            self.flow_view.ignore_key_event = True


class FlowViewProxyHoverWidget(FlowViewProxyWidget):
    """Additional hover controls for QProxyWidgets in the flow."""

    def __init__(self, flow_view: FlowView, parent=None):
        super(FlowViewProxyHoverWidget, self).__init__(flow_view, parent)
        self._is_hovered: bool = False

    def hoverEnterEvent(self, event: QGraphicsSceneHoverEvent):
        super().hoverEnterEvent(event)
        self._is_hovered = True

    def hoverLeaveEvent(self, event: QGraphicsSceneHoverEvent):
        super().hoverLeaveEvent(event)
        self._is_hovered = False

    def is_hovered(self) -> bool:
        return self._is_hovered



# import math
from typing import List, Any, Tuple
from qtpy.QtCore import QPointF, Qt
from qtpy.QtGui import QPainter, QColor, QRadialGradient, QPainterPath, QPen
from qtpy.QtWidgets import (
    QGraphicsPathItem,
    QGraphicsItem,
    QStyleOptionGraphicsItem,
    QGraphicsEllipseItem,
)

from ...GUIBase import GUIBase
from ...utils import sqrt
from ...utils import pythagoras
from ...flows.nodes.PortItem import PortItem
from .ConnectionAnimation import ConnPathItemsAnimation, ConnPathItemsAnimationScaled

from ryvencore_qt.src.Design import Design
from ryvencore.NodePort import NodeInput, NodeOutput


class ConnectionItem(GUIBase, QGraphicsPathItem):
    """The GUI representative for a connection. The classes ExecConnectionItem and DataConnectionItem will be ready
    for reimplementation later, so users can add GUI for the enhancements of DataConnection and ExecConnection,
    like input fields for weights."""

    def __init__(self, connection: Tuple[NodeOutput, NodeInput], session_design: Design):
        QGraphicsPathItem.__init__(self)

        self.setAcceptHoverEvents(True)

        self.connection = connection
        out, inp = self.connection

        out_port_index = out.node.outputs.index(out)
        inp_port_index = inp.node.inputs.index(inp)
        assert hasattr(out.node, 'gui') and hasattr(inp.node, 'gui')
        self.out_item: PortItem = out.node.gui.item.outputs[out_port_index]
        self.inp_item: PortItem = inp.node.gui.item.inputs[inp_port_index]

        self.session_design = session_design
        self.session_design.flow_theme_changed.connect(self.recompute)
        self.session_design.performance_mode_changed.connect(self.recompute)

        # for rendering flow pictures
        self.setCacheMode(QGraphicsItem.DeviceCoordinateCache)
        self.setFlag(QGraphicsItem.ItemIsSelectable)

        diam = 12.5
        self.num_dots = 40
        self.dots = [
            QGraphicsEllipseItem(-diam / 2, -diam / 2, diam, diam, self) 
            for _ in range(self.num_dots)
        ]
        for dot in self.dots:
            dot.setVisible(False)

        # TODO: the connection animation is currently unused because we need a 
        #       signel for when the node output is updated
        self.items_path_animation = ConnPathItemsAnimation(self.dots, self)
        self.connection_animation = ConnPathItemsAnimationScaled(self.items_path_animation)

        self.recompute()

    def __str__(self):
        out, inp = self.connection
        node_in_name = f'{inp.node.gui.item}'
        node_in_index = inp.node.inputs.index(inp)
        node_out_name = f'{out.node.gui.item}'
        node_out_index = out.node.outputs.index(out)
        return f'{node_out_index}->{node_in_index} ({node_out_name}, {node_in_name})'

    def paint(self, painter: QPainter, option: QStyleOptionGraphicsItem, widget = None):
        # draw path
        painter.setBrush(self.brush())
        painter.setPen(self.pen())
        painter.drawPath(self.path())
        # return super().paint(painter, option, widget)

    def recompute(self) -> None:
        """Updates scene position and recomputes path, pen, gradient and dots"""

        # dots
        self.items_path_animation.recompute()

        # position
        self.setPos(self.out_pos())

        # path
        p1: QPointF = QPointF(self.out_item.pin.width_no_padding() * 0.5, 0)
        p2: QPointF = self.inp_pos() - self.scenePos() - QPointF(self.inp_item.pin.width_no_padding() * 0.5, 0)
        self.setPath(self.connection_path(p1, p2))

        # pen
        pen: QPen = self.get_pen()

        # brush
        self.setBrush(Qt.NoBrush)

        #   gradient
        if self.session_design.performance_mode == 'pretty':
            c = pen.color()
            w = self.path().boundingRect().width()
            h = self.path().boundingRect().height()
            gradient = QRadialGradient(
                self.boundingRect().center(), 
                pythagoras(w, h) / 2
            )

            c_r: int = c.red()
            c_g: int = c.green()
            c_b: int = c.blue()

            # this offset will be 1 if inp.x >> out.x and 0 if inp.x < out.x
            # hence, no fade for the gradient if the connection goes backwards
            offset_mult: float = max(
                0,
                min(
                    (self.inp_pos().x() - self.out_pos().x()) / 200,
                    1
                )
            )

            # and if the input is very far away from the output, decrease the gradient fade so the connection
            # doesn't fully disappear at the ends and stays visible
            if self.inp_pos().x() > self.out_pos().x():
                offset_mult = min(
                    offset_mult,
                    2000 / (self.dist(self.inp_pos(), self.out_pos()))
                )
                # zucker.

            gradient.setColorAt(0.0, QColor(c_r, c_g, c_b, 255))
            gradient.setColorAt(0.75, QColor(c_r, c_g, c_b, 255 - round(55 * offset_mult)))
            gradient.setColorAt(0.95, QColor(c_r, c_g, c_b, 255 - round(255 * offset_mult)))

            pen.setBrush(gradient)

        self.setPen(pen)

    def out_pos(self) -> QPointF:
        """The current global scene position of the pin of the output port"""

        return self.out_item.pin.get_scene_center_pos()

    def inp_pos(self) -> QPointF:
        """The current global scene position of the pin of the input port"""

        return self.inp_item.pin.get_scene_center_pos()

    def itemChange(self, change, value):
        if change == QGraphicsItem.ItemSelectedHasChanged or (
            change == QGraphicsItem.ItemVisibleHasChanged and self.isVisible()
        ):
            self.set_highlighted(self.isSelected())
        return QGraphicsItem.itemChange(self, change, value)

    def set_highlighted(self, b: bool):
        pen: QPen = self.pen()

        if b:
            pen.setWidthF(self.pen_width() * 2)
        else:
            pen.setWidthF(self.pen_width())
            self.recompute()

        self.setPen(pen)

    def get_pen(self) -> QPen:
        raise NotImplementedError

    def pen_width(self) -> int:
        raise NotImplementedError

    def get_style_color(self):
        raise NotImplementedError

    def flow_theme(self):
        return self.session_design.flow_theme

    def hoverEnterEvent(self, event):
        self.set_highlighted(True)
        super().hoverEnterEvent(event)

    def hoverLeaveEvent(self, event):
        if not self.isSelected():
            self.set_highlighted(False)
        super().hoverLeaveEvent(event)

    # uncomment to test connection animation:
    # 
    # def mouseReleaseEvent(self, event) -> None:
    #    self.connection_animation.toggle()
        
    @staticmethod
    def dist(p1: QPointF, p2: QPointF) -> float:
        """Returns the diagonal distance between the points using pythagoras"""

        dx = p2.x() - p1.x()
        dy = p2.y() - p1.y()
        return sqrt((dx**2) + (dy**2))

    @staticmethod
    def connection_path(p1: QPointF, p2: QPointF) -> QPainterPath:
        """Returns the painter path for drawing the connection, using the usual cubic connection path by default"""

        return default_cubic_connection_path(p1, p2)


class ExecConnectionItem(ConnectionItem):
    def pen_width(self):
        return self.flow_theme().exec_conn_width

    def get_pen(self):
        theme = self.flow_theme()
        pen = QPen(theme.exec_conn_color, theme.exec_conn_width)
        pen.setStyle(theme.exec_conn_pen_style)
        pen.setCapStyle(Qt.RoundCap)
        return pen

    def get_style_color(self):
        return self.flow_theme().exec_conn_color


class DataConnectionItem(ConnectionItem):
    def pen_width(self):
        return self.flow_theme().data_conn_width

    def get_pen(self):
        theme = self.flow_theme()
        pen = QPen(theme.data_conn_color, theme.data_conn_width)
        pen.setStyle(theme.data_conn_pen_style)
        pen.setCapStyle(Qt.RoundCap)
        return pen

    def get_style_color(self):
        return self.flow_theme().data_conn_color


def default_cubic_connection_path(p1: QPointF, p2: QPointF) -> QPainterPath:
    """Returns the nice looking QPainterPath from p1 to p2"""

    path = QPainterPath()

    path.moveTo(p1)

    dx = p2.x() - p1.x()
    adx = abs(dx)
    dy = p2.y() - p1.y()
    ady = abs(dy)
    distance = sqrt((dx**2) + (dy**2))
    x1, y1 = p1.x(), p1.y()
    x2, y2 = p2.x(), p2.y()

    if ((x1 < x2 - 30) or distance < 100) and (x1 < x2):
        # STANDARD FORWARD
        path.cubicTo(x1 + ((x2 - x1) / 2), y1,
                     x1 + ((x2 - x1) / 2), y2,
                     x2, y2)
    elif x2 < x1 - 100 and adx > ady * 2:
        # STRONG BACKWARDS
        path.cubicTo(x1 + 100 + (x1 - x2) / 10, y1,
                     x1 + 100 + (x1 - x2) / 10, y1 + (dy / 2),
                     x1 + (dx / 2), y1 + (dy / 2))
        path.cubicTo(x2 - 100 - (x1 - x2) / 10, y2 - (dy / 2),
                     x2 - 100 - (x1 - x2) / 10, y2,
                     x2, y2)
    else:
        # STANDARD BACKWARDS
        path.cubicTo(x1 + 100 + (x1 - x2) / 3, y1,
                     x2 - 100 - (x1 - x2) / 3, y2,
                     x2, y2)

    return path



from qtpy.QtWidgets import QWidget, QHBoxLayout, QLabel, QMenu, QAction
from qtpy.QtGui import QIcon, QImage, QMouseEvent
from qtpy.QtCore import Qt, QEvent, QBuffer, QByteArray, QIODevice

from ..GlobalAttributes import Location
from .ListWidget_NameLineEdit import ListWidget_NameLineEdit


class FlowsList_FlowWidget(QWidget):
    """A QWidget representing a single Flow for the FlowsListWidget."""

    def __init__(self, flows_list_widget, session_gui, flow):
        super().__init__()

        self.session_gui = session_gui
        self.flow = flow
        self.flow_view = self.session_gui.flow_views[flow]
        self.flows_list_widget = flows_list_widget
        self.previous_flow_title = ''
        self._thumbnail_source = ''
        self.ignore_title_line_edit_signal = False


        # UI

        main_layout = QHBoxLayout()
        main_layout.setContentsMargins(0, 0, 0, 0)

        #   create icon

        # TODO: change this icon
        flow_icon = QIcon(Location.PACKAGE_PATH + '/resources/pics/script_picture.png')

        icon_label = QLabel()
        icon_label.setFixedSize(20, 20)
        icon_label.setStyleSheet('border:none;')
        icon_label.setPixmap(flow_icon.pixmap(20, 20))
        main_layout.addWidget(icon_label)

        #   title line edit

        self.title_line_edit = ListWidget_NameLineEdit(flow.title, self)
        self.title_line_edit.setPlaceholderText('title')
        self.title_line_edit.setEnabled(False)
        self.title_line_edit.editingFinished.connect(self.title_line_edit_editing_finished)

        main_layout.addWidget(self.title_line_edit)

        self.setLayout(main_layout)



    def mouseDoubleClickEvent(self, event: QMouseEvent):
        if event.button() == Qt.LeftButton:
            if self.title_line_edit.geometry().contains(event.pos()):
                self.title_line_edit_double_clicked()
                return


    def event(self, event: QEvent):
        if event.type() == QEvent.ToolTip:

            # generate preview img as QImage
            img: QImage = self.flow_view.get_viewport_img().scaledToHeight(200)

            # store the img data in QBuffer to load it directly from memory
            buffer = QBuffer()
            img.save(device=buffer, format='PNG')  # type: ignore
            # QBuffer inherits QIODevice, but mypy doesn't know that

            # generate html from data in memory
            html = f"<img src='data:image/png;base64, { bytes( buffer.data().toBase64() ).decode() }'>"

            # show tooltip
            self.setToolTip(html)

        return QWidget.event(self, event)


    def contextMenuEvent(self, event: QEvent):
        menu: QMenu = QMenu(self)

        delete_action = QAction('delete')
        delete_action.triggered.connect(self.action_delete_triggered)

        actions = [delete_action]
        for a in actions:
            menu.addAction(a)

        menu.exec_(event.globalPos())


    def action_delete_triggered(self):
        self.flows_list_widget.del_flow(self.flow, self)


    def title_line_edit_double_clicked(self):
        self.title_line_edit.setEnabled(True)
        self.title_line_edit.setFocus()
        self.title_line_edit.selectAll()

        self.previous_flow_title = self.title_line_edit.text()


    def title_line_edit_editing_finished(self):
        if self.ignore_title_line_edit_signal:
            return

        title = self.title_line_edit.text()

        self.ignore_title_line_edit_signal = True

        if self.session_gui.core_session.flow_title_valid(title):
            self.session_gui.core_session.rename_flow(flow=self.flow, title=title)
        else:
            self.title_line_edit.setText(self.previous_flow_title)

        self.title_line_edit.setEnabled(False)
        self.ignore_title_line_edit_signal = False



import bisect
import enum
import json
import pathlib
from math import sqrt
from typing import List, Dict

from qtpy.QtCore import QPointF, QByteArray

from ryvencore.utils import serialize, deserialize
from .GlobalAttributes import *

def generate_name(obj, name):
   return f'{name}:[{id(obj)}]'
    
def pythagoras(a, b):
    return sqrt(a ** 2 + b ** 2)


def get_longest_line(s: str):
    lines = s.split('\n')
    lines = [line.replace('\n', '') for line in lines]
    longest_line_found = ''
    for line in lines:
        if len(line) > len(longest_line_found):
            longest_line_found = line
    return line


def shorten(s: str, max_chars: int, line_break: bool = False):
    """Ensures, that a given string does not exceed a given max length. If it would, its cut in the middle."""
    l = len(s)
    if l > max_chars:
        insert = ' . . . '
        if line_break:
            insert = '\n'+insert+'\n'
        insert_length = len(insert)
        left = s[:round((max_chars-insert_length)/2)]
        right = s[round(l-((max_chars-insert_length)/2)):]
        return left+insert+right
    else:
        return s


def pointF_mapped(p1, p2):
    """adds the floating part of p2 to p1"""
    p2.setX(p1.x() + p2.x()%1)
    p2.setY(p1.y() + p2.y()%1)
    return p2

def points_dist(p1, p2):
    return sqrt(abs(p1.x() - p2.x())**2 + abs(p1.y() - p2.y())**2)

def middle_point(p1, p2):
    return QPointF((p1.x() + p2.x())/2, (p1.y() + p2.y())/2)


class MovementEnum(enum.Enum):
    # this should maybe get removed later
    mouse_clicked = 1
    position_changed = 2
    mouse_released = 3


def get_resource(filepath: str):
    return pathlib.Path(Location.PACKAGE_PATH, 'resources', filepath)


def change_svg_color(filepath: str, color_hex: str):
    """Loads an SVG, changes all '#xxxxxx' occurrences to color_hex, renders it into and a pixmap and returns it"""

    # https://stackoverflow.com/questions/15123544/change-the-color-of-an-svg-in-qt

    from qtpy.QtSvg import QSvgRenderer
    from qtpy.QtGui import QPixmap, QPainter
    from qtpy.QtCore import Qt

    with open(filepath) as f:
        data = f.read()
    data = data.replace('fill:#xxxxxx', 'fill:'+color_hex)

    svg_renderer = QSvgRenderer(QByteArray(bytes(data, 'ascii')))

    pix = QPixmap(svg_renderer.defaultSize())
    pix.fill(Qt.transparent)
    pix_painter = QPainter(pix)
    svg_renderer.render(pix_painter)

    return pix



import textdistance


def dec(i: int, length: int) -> int:
    if i != 0:
        return i - 1
    else:
        return length - 1

def inc(i: int, length: int) -> int:
    if i != length - 1:
        return i + 1
    else:
        return 0


def sort_nodes(nodes):
    return sorted(nodes, key=lambda x: x.title.lower())


def sort_by_val(d: dict) -> dict:
    return {
        k: v
        for k, v in sorted(
            d.items(),

            # x: (key, value); sort by value
            key=lambda x: x[1]
        )
    }


def search(items: dict, text: str) -> dict:
    """performs the search on `items` under search string `text`"""
    dist = textdistance.sorensen_dice.distance

    distances = {}

    for item, tags in items.items():
        min_dist = 1.0
        for tag in tags:
            min_dist = min(min_dist, dist(text, tag))

        distances[item] = min_dist

    return sort_by_val(distances)



import json

from qtpy.QtWidgets import QLineEdit, QWidget, QLabel, QGridLayout, QHBoxLayout, QSpacerItem, QSizePolicy, QStyleOption, QStyle
from qtpy.QtGui import QFont, QPainter, QColor, QDrag
from qtpy.QtCore import Signal, Qt, QMimeData, QByteArray


class NodeWidget(QWidget):

    chosen = Signal()
    custom_focused_from_inside = Signal()

    @staticmethod
    def _create_mime_data(node) -> QMimeData:
        mime_data = QMimeData()
        mime_data.setData('application/json', QByteArray(bytes(json.dumps(
                {
                    'type': 'node',
                    'node identifier': node.identifier,
                }
            ), encoding='utf-8')))
        return mime_data
    
    def __init__(self, parent, node):
        super(NodeWidget, self).__init__(parent)

        self.custom_focused = False
        self.node = node

        self.left_mouse_pressed_on_me = False

        # UI
        main_layout = QGridLayout()
        main_layout.setContentsMargins(0, 0, 0, 0)

        self_ = self
        class NameLabel(QLineEdit):
            def __init__(self, text):
                super().__init__(text)

                self.setReadOnly(True)
                self.setFont(QFont('Source Code Pro', 8))
            def mouseMoveEvent(self, ev):
                self_.custom_focused_from_inside.emit()
                ev.ignore()
            def mousePressEvent(self, ev):
                ev.ignore()
            def mouseReleaseEvent(self, ev):
                ev.ignore()

        name_label = NameLabel(node.title)

        type_layout = QHBoxLayout()

        #type_label = QLabel(node.type_)
        #type_label.setFont(QFont('Segoe UI', 8, italic=True))
        # type_label.setStyleSheet('color: white;')

        main_layout.addWidget(name_label, 0, 0)
        #main_layout.addWidget(type_label, 0, 1)

        self.setLayout(main_layout)
        self.setContentsMargins(0, 0, 0, 0)
        self.setMaximumWidth(250)

        self.setToolTip(node.__doc__)
        self.update_stylesheet()


    def mousePressEvent(self, event):
        self.custom_focused_from_inside.emit()
        if event.button() == Qt.LeftButton:
            self.left_mouse_pressed_on_me = True

    def mouseMoveEvent(self, event):
        if self.left_mouse_pressed_on_me:
            drag = QDrag(self)
            mime_data = QMimeData()
            mime_data.setData('application/json', bytes(json.dumps(
                {
                    'type': 'node',
                    'node identifier': self.node.identifier,
                }
            ), encoding='utf-8'))
            drag.setMimeData(mime_data)
            drop_action = drag.exec_()

    def mouseReleaseEvent(self, event):
        self.left_mouse_pressed_on_me = False
        if self.geometry().contains(self.mapToParent(event.pos())):
            self.chosen.emit()

    def set_custom_focus(self, new_focus):
        self.custom_focused = new_focus
        self.update_stylesheet()

    def update_stylesheet(self):
        color = self.node.GUI.color if hasattr(self.node, 'GUI') else '#888888'

        r, g, b = QColor(color).red(), QColor(color).green(), QColor(color).blue()

        new_style_sheet = f'''
NodeWidget {{
    border: 1px solid rgba(255,255,255,150);
    border-radius: 2px;
    {(
        f'background-color: rgba(255,255,255,80);'
    ) if self.custom_focused else ''}
}}
QLabel {{
    background: transparent;
}}
QLineEdit {{
    background: transparent;
    border: none;
    padding: 2px;
}}
        '''

        self.setStyleSheet(new_style_sheet)

    def paintEvent(self, event):  # just to enable stylesheets
        o = QStyleOption()
        o.initFrom(self)
        p = QPainter(self)
        self.style().drawPrimitive(QStyle.PE_Widget, o, p, self)


# prevent circular imports
from __future__ import annotations
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from .NodeItem import NodeItem

from qtpy.QtCore import QObject, QPropertyAnimation, Property, QParallelAnimationGroup, QTimeLine
from qtpy.QtGui import QColor
from qtpy.QtWidgets import QGraphicsItem

class NodeItemAnimator(QObject):

    def __init__(self, node_item: NodeItem):
        super(NodeItemAnimator, self).__init__()

        self.node_item: QGraphicsItem = node_item
        self.animation_running = False

        # title color
        self.title_activation_animation = QPropertyAnimation(self, b"p_title_color")  # type: ignore
        self.title_activation_animation.setDuration(700)
        # body color
        self.body_activation_animation = QPropertyAnimation(self, b"p_body_color")  # type: ignore
        self.body_activation_animation.setDuration(700)
        # transform
        self.scale_animation = QPropertyAnimation(self.node_item, b'scale')  # type: ignore
        self.scale_animation.setDuration(700)
        self.scalar = 1.04
        
        self.animation = QParallelAnimationGroup()
        self.animation.addAnimation(self.title_activation_animation)
        self.animation.addAnimation(self.body_activation_animation)
        self.animation.addAnimation(self.scale_animation)
        self.animation.finished.connect(self.finished)

    def start(self):
        self.animation_running = True
        self.animation.start()

    def stop(self):
        # reset color values. it would just freeze without
        self.title_activation_animation.setCurrentTime(self.title_activation_animation.duration())
        self.body_activation_animation.setCurrentTime(self.body_activation_animation.duration())
        self.scale_animation.setCurrentTime(self.scale_animation.duration())
        
        self.animation.stop()

    def finished(self):
        self.animation_running = False

    def running(self):
        return self.animation_running

    def reload_values(self):
        self.stop()

        # self.node_item.title_label.update_design()
        self.title_activation_animation.setKeyValueAt(0, self.get_title_color())
        self.title_activation_animation.setKeyValueAt(0.3, self.get_body_color().lighter().lighter())
        self.title_activation_animation.setKeyValueAt(1, self.get_title_color())

        self.body_activation_animation.setKeyValueAt(0, self.get_body_color())
        self.body_activation_animation.setKeyValueAt(0.3, self.get_body_color().lighter())
        self.body_activation_animation.setKeyValueAt(1, self.get_body_color())
        
        self.scale_animation.setKeyValueAt(0, 1)
        self.scale_animation.setKeyValueAt(0.3, self.scalar)
        self.scale_animation.setKeyValueAt(1, 1)

    def fading_out(self):
        return self.title_activation_animation.currentTime()/self.title_activation_animation.duration() >= 0.3

    def set_animation_max(self):
        self.title_activation_animation.setCurrentTime(0.3*self.title_activation_animation.duration())
        self.body_activation_animation.setCurrentTime(0.3*self.body_activation_animation.duration())

    # BODY COLOR
    def get_body_color(self):
        return self.node_item.color

    def set_body_color(self, val):
        self.node_item.color = val
        QGraphicsItem.update(self.node_item)

    p_body_color = Property(QColor, get_body_color, set_body_color)

    # TITLE COLOR
    def get_title_color(self):
        return self.node_item.widget.title_label.color

    def set_title_color(self, val):
        self.node_item.widget.title_label.color = val
        # QGraphicsItem.update(self.node_item)

    p_title_color = Property(QColor, get_title_color, set_title_color)